/**
*
*/
package owrti;
import owrti.evt.*;
import RTIcap.*;
import npacrti.*;

import java.util.*;

/*
*
* @date April 20, 1998 ( Last Updated March 31, 1999)
* @author H. Timucin Ozdemir
*
*/
public class 
DeclarationManager extends _DeclarationManagerInterfaceImplBase
                   implements RTIDebug
{
    FederationExecution fedEx;
    ObjectClassNaming objNaming ;
    InteractionNaming interNaming ;
    RTIMsgDispatcher  dispatcher = new RTIMsgDispatcher();

    /** */
    public DeclarationManager( ObjectClassNaming _objNaming, 
                        InteractionNaming _interNaming, 
                        FederationExecution _fedEx) { 
       objNaming = _objNaming;
       interNaming = _interNaming;
       fedEx = _fedEx;
    }// end of CONSTRUCTOR

    /**
    * 5.2
    */
    public void 
    publishObjectClass(short theClass, short[] attributeList, short _fedId   )
         throws RTIcap.ObjectClassNotDefined, RTIcap.AttributeNotDefined, 
                RTIcap.FederateNotExecutionMember, RTIcap.RTIinternalError 
    {   isExecutionMember( _fedId );
        if(DEBUG) System.out.println(" in publishObjectClass() ");
        ObjectRecord objRecord = (ObjectRecord) objNaming.getObjectRecord(theClass);
        objRecord.addFederate(_fedId);
        if(DEBUG) System.out.println(" Object Exists! () ");
        AttributeWatcher attWatcher ; 
        boolean except = false;
        for(int i=0;i<attributeList.length;i++) {
           attWatcher = (AttributeWatcher) objRecord.getAttributeWatcher(attributeList[i]);
           if(attWatcher!=null) attWatcher.addFederate(fedEx.getFederateInfo(_fedId));
           else except = true;
        }// end of for
        if(except) throw new RTIcap.AttributeNotDefined();
        dispatcher.addRTIMsg(new BroadcastStartUpdatesMessage(theClass,attributeList,_fedId));
        if(DEBUG) System.out.println(" out of publishObjectClass() ");
    }// end of publishObjectClass()

    /**
    * 5.3
    */
    public void 
    unpublishObjectClass(short theClass, short _fedId )
         throws RTIcap.ObjectClassNotDefined, RTIcap.FederateOwnsAttributes, 
                RTIcap.FederateNotExecutionMember, RTIcap.RTIinternalError
    {  if(DEBUG) System.out.println(" in unpublishObjectClass() ");
       isExecutionMember( _fedId );
       ObjectRecord objRecord = (ObjectRecord) objNaming.getObjectRecord(theClass);
       objRecord.removeFederate(_fedId);

       AttributeWatcher attWatcher ; 
       boolean except = false;
       short[] attributeList = objRecord.getAllAttributeHandles();
       for(int i=0;i<attributeList.length;i++) {
          attWatcher = (AttributeWatcher) objRecord.getAttributeWatcher(attributeList[i]);
          if(attWatcher!=null) attWatcher.removeFederate(_fedId);
          else except = true;
       }// end of for
       if(DEBUG) System.out.println(" out of unpublishObjectClass() ");
    }// end of unpublishObjectClass() 

    /**
    * 5.4
    */
    public void 
    publishInteractionClass(short theInteraction, short _fedId   )
         throws RTIcap.InteractionClassNotDefined, 
                RTIcap.FederateNotExecutionMember, RTIcap.RTIinternalError 
    {  if(DEBUG) System.out.println(" in publishInteractionClass() ");
       isExecutionMember( _fedId );
       InteractionRecord rec = interNaming.getInteractionRecord(theInteraction);
       FederateInfo fedInf = fedEx.getFederateInfo(_fedId);
       if(fedInf!=null) {
          rec.addFederate(fedEx.getFederateInfo(_fedId));
          int handle = fedInf.getGlobalEvtRepository().addEvent(new EvtStartInteractionGeneration(theInteraction,_fedId));
          fedInf.addEventInfo(new EventInfo( handle, EventInfo.LOCAL ));
       }// end of if
       if(DEBUG) System.out.println(" out of publishInteractionClass() ");
    }// end of publishInteractionClass()

    /**
    * 5.5
    */
    public void 
    unpublishInteractionClass(short theInteraction, short _fedId )
         throws RTIcap.InteractionClassNotDefined, 
                RTIcap.FederateNotExecutionMember, RTIcap.RTIinternalError 
    {  if(DEBUG) System.out.println(" out of unpublishInteractionClass() ");
       isExecutionMember( _fedId );
       InteractionRecord rec = interNaming.getInteractionRecord(theInteraction);
       
       FederateInfo fedInf = fedEx.getFederateInfo(_fedId);
       if(fedInf!=null) {
         rec.removeFederate(_fedId);
         int handle = fedInf.getGlobalEvtRepository().addEvent(new EvtStopInteractionGeneration(theInteraction,_fedId));
         fedInf.addEventInfo(new EventInfo( handle, EventInfo.LOCAL ));
       }// end of if

       if(DEBUG) System.out.println(" out of unpublishInteractionClass() ");
    }// end of unpublishInteractionClass()

    /**
    * 5.6
    *
    *  Get ObjectRecord
    *  Find AttributeWatcher Object
    *  Subscribe yourself to this queue.
    *
    */
    public void 
    subscribeObjectClassAttributes( short theClass, short[] attributeList, 
                                    RTIcap._Boolean active, short _fedId   )
         throws RTIcap.ObjectClassNotDefined, RTIcap.AttributeNotDefined, 
                RTIcap.FederateNotExecutionMember, RTIcap.RTIinternalError 
    {  int i;
       if(DEBUG) System.out.println(" in subscribeObjectClassAttribute() ");
       isExecutionMember( _fedId );
       ObjectRecord objRecord = (ObjectRecord) objNaming.getObjectRecord(theClass);
       /**************************************************************/
       /** Inform federate about previously defined object instances */
       /**************************************************************/
       objRecord.broadcastObjectInstances(fedEx.getFederateInfo(_fedId));
       for(i=0;i<attributeList.length;i++) 
           objRecord.haveAttribute(attributeList[i]);

       AttributeWatcher attWatcher ; 
       boolean except = false;
       String mess = new String("Attributes ");

       for(i=0;i<attributeList.length;i++) {
           attWatcher = (AttributeWatcher) objRecord.getAttributeWatcher(attributeList[i]);
           if(attWatcher!=null) 
              attWatcher.register_subscriber(fedEx.getFederateInfo(_fedId));
           else { except = true;
             mess += (attributeList[i]+", ");
             if(DEBUG) System.out.println(attributeList[i]+" does not exist! ");
           }// end of else 
        }// end of for
        if(DEBUG) System.out.println(" out of subscribeObjectClassAttribute() ");
        if(except) throw new RTIcap.AttributeNotDefined(0,mess+" are not known");
    }// end of subscribeObjectClassAttribute()

    /**
    * 5.7
    */
    public void 
    unsubscribeObjectClassAttributes( short theClass, short[] attributeList,
                                      short _fedId   )
         throws RTIcap.ObjectClassNotDefined, RTIcap.FederateNotExecutionMember, 
                RTIcap.RTIinternalError 
    {   
        isExecutionMember( _fedId );

        ObjectRecord objRecord = (ObjectRecord) objNaming.getObjectRecord(theClass);
        AttributeWatcher attWatcher ; 
        boolean except = false;
        String mess = new String("Attributes ");
        for(int i=0;i<attributeList.length;i++) {
           attWatcher = (AttributeWatcher) 
                      objRecord.getAttributeWatcher(attributeList[i]);
           if(attWatcher!=null) attWatcher.deregister_subscriber(_fedId);
           else { except = true; mess += (attributeList[i]+", "); }
        }// end of for

        if(except) throw new RTIcap.RTIinternalError(0,mess+" are not known");
    }// end of unsubscribeObjectClassAttributes()

    /**
    * 5.7
    */
    public void unsubscribeObjectClass( short theClass, short _fedId   )
           throws RTIcap.ObjectClassNotDefined, RTIcap.RTIinternalError,
                  RTIcap.FederateNotExecutionMember { 
      isExecutionMember(_fedId);
      ObjectRecord rec = (ObjectRecord) objNaming.getObjectRecord(theClass);
      unsubscribeObjectClassAttributes(theClass,rec.getAllAttributeHandles(),_fedId);
    }// end of unsubscribeObjectClass()

    /**
    * 5.8
    */
    public void 
    subscribeInteractionClass(short theClass,RTIcap._Boolean active,short _fedId   )
         throws RTIcap.InteractionClassNotDefined, 
                RTIcap.FederateNotExecutionMember, RTIcap.RTIinternalError 
    {  if(DEBUG) System.out.println(" in subscribeInteractionClass() ");
       isExecutionMember( _fedId );
       InteractionRecord rec = interNaming.getInteractionRecord(theClass);
       //if(rec!=null) rec.register_subscriber(_fedHandle,_fedId);
       if(rec!=null) rec.register_subscriber(fedEx.getFederateInfo(_fedId));
       if(DEBUG) System.out.println(" out of subscribeInteractionClass() ");
    }// end of subscribeInteractionClass()

    /**
    * 5.9
    */
    public void 
    unsubscribeInteractionClass( short theClass, short _fedId   )
         throws RTIcap.InteractionClassNotDefined, 
                RTIcap.FederateNotExecutionMember, RTIcap.RTIinternalError 
    { isExecutionMember( _fedId );
      InteractionRecord irec = interNaming.getInteractionRecord(theClass);
      irec.deregister_subscriber(_fedId);
    }// end of unsubscribeInteractionClass()

    /** */
    final void 
    isExecutionMember( short fedId ) throws RTIcap.FederateNotExecutionMember {
      if(fedEx.getFederateInfo(fedId)==null) 
         throw new RTIcap.FederateNotExecutionMember();
    }// end of isExecutionMember() 

    //
    // if there is more than 1 federate in the federation then 
    // send startUpdate message to the federate. If this federate is
    // the second federate, then we need to send the same same message 
    // to the first federate.
    //
    class BroadcastStartUpdatesMessage implements RTIMessage, RTIDebug {
       short theClass;
       short[] attributeList;
       short _fedId;
       BroadcastStartUpdatesMessage(short theClass, short[] attributeList, 
                                    short _fedId) {
         this.theClass=theClass;
         this.attributeList=attributeList;
         this._fedId=_fedId;
       }// end of CONSTRUCTOR

       public void execute() {
        FederateInfo fedInf = fedEx.getFederateInfo(_fedId);
        int numFeds = fedEx.getNumberOfFederates();
        if(fedInf!=null&&numFeds>1) {
            if(numFeds==2) {
               Enumeration enu = fedEx.getAllFederateInfos();
               FederateInfo fedInfoItem;
               for(;enu.hasMoreElements();) {
                  fedInfoItem = (FederateInfo) enu.nextElement();
                  if(fedInfoItem!=null) {
                    int handle = fedInfoItem.getGlobalEvtRepository().addEvent(new EvtStartUpdate(theClass,attributeList,fedInfoItem.getFederateID()));
                    fedInfoItem.addEventInfo(new EventInfo( handle, EventInfo.LOCAL ));
                  }// end of if

               }// end of for
            }
            else {
              int handle = fedInf.getGlobalEvtRepository().addEvent(new EvtStartUpdate(theClass,attributeList,_fedId));
              fedInf.addEventInfo(new EventInfo( handle, EventInfo.LOCAL ));
            }// end of else
        }// end of if
       }// end of execute

    }// end of class BroadcastStartUpdatesMessage()

}// end of DeclarationManager

