package owrti;

import RTIcap.*;
import java.util.*;
import java.io.*;

/**
*
* This object keeps the Receive Order Event Queue.
*
* <dl>
* <dt> Date: February 7, 1999 (timucin)
* <dd> Asynchronous delivery option forks a thread watching the event queue.
* Whenever something is added, then thread starts sending them to the 
* federate without waiting federate's request (tick(),timeAdvance(), etc).
*
* </dl>
*
* @date April 20, 1997 ( Updated Fubruary 7, 1999 )
* @author H. Timucin Ozdemir
*
*/
public class EvtQueueRO extends EvtQueue implements RTIDebug, Runnable 
{

   Vector evts = new Vector();
   Hashtable idToEvtInfo = new Hashtable();

   protected RTIcap.FederateAmbassador fedAmb;
   protected RTIEventRepository que;
   protected FederateInfo fedInf;
   protected String FederateName ;
   long      numberOfEvents = 0;

   Thread    runner = null ;
   boolean   run_stay = false ;
   boolean   aDelivery = false ;

   /**
   */
   public 
   EvtQueueRO( FederateInfo _fed, RTIEventRepository _que )
   {
      orderType = ReceiveOrder;
      fedAmb = _fed.getFederateAmbassador();
      fedInf = _fed;
      FederateName = fedInf.getFederateName();
      que = _que;

   }// end of CONSTRUCTOR

   /**
   * Executes  all the events in the EventQue
   */
   public synchronized int processEvents()
   {

      return(fedAmb!=null?processEvents(fedAmb):0);

   }// end of processEvents()

   /**
   * Executes  all the events in the EventQue
   */
   protected synchronized int processEvents(RTIcap.FederateAmbassador _fed)
   {

      int   num_of_exec = 0;
      EventInfo evtInf;
      RTIEvent evtInstance;
      Enumeration e ;

      synchronized( evts ) { e = evts.elements(); }

      if(!e.hasMoreElements() && DEBUG) 
        System.out.println("There is nothing to process for "+FederateName+"!");
      for(;e.hasMoreElements()&&num_of_exec<10;)
      {
          evtInf = (EventInfo) e.nextElement();
          if(evtInf!=null) 
          { 
            evtInstance = null;

            /** Find the RTIEvent object */
            if(que!=null) evtInstance = que.getRTIEvent(evtInf.getHandle());

            if(evtInstance!=null) 
            {
               log(" Execute ",evtInstance);

               // Note: We need a way to inform that this federate is no
               // longer in the federation execution. Question is
               // can we conclude that this federate is not responsive 
               // from the first exception or should we try to do more
               // fancy things. We can anounce that this federate
               // left the building as soon as we catch the first exception
               // about federateAmbassador when it is not responsing. 
               /** Execute this event */
               try{
               if(_fed!=null) evtInstance.execute(_fed);
               } catch(FederateNotResponding fnr) 
                 { fedInf.getFederationExecution().resignFederateFromExecution(fedInf.getFederateID()); }
               num_of_exec++;
 
               /** Delete this record from the Event List */
               synchronized( evts ) { evts.removeElement(evtInf); }

               /** Delete this record from the Event List */
               synchronized( idToEvtInfo )
               { idToEvtInfo.remove(new Integer(evtInf.getHandle())); }

               numberOfEvents--;
 
            }// end of if 

            evtInf = null;

          }// end of if

      }// end of for

      e = null;

      return num_of_exec;

   }// end of processEvents()

   /**
   */
   public synchronized void 
   releaseEvents()
   {
       synchronized( evts ) { evts.removeAllElements(); }
       synchronized( idToEvtInfo ) { idToEvtInfo.clear(); } 

   }// end of releaseEvents()

   /**
   */
   public synchronized void 
   addEventInfo(EventInfo evtInfo)
   {
      if(numberOfEvents<3000)
      {
         synchronized( evts ) { evts.addElement(evtInfo); }
         synchronized( idToEvtInfo ) 
         { idToEvtInfo.put(new Integer(evtInfo.getHandle()),evtInfo); }

         numberOfEvents++;

         //if(aDelivery) notify();

      }// end of if

   }// end of addEventInfo

   /**
   */
   public synchronized void
   cancelEvent(int evtID)
   {
      EventInfo evtInfo = null;
      synchronized( idToEvtInfo ) 
      { evtInfo = (EventInfo) idToEvtInfo.get(new Integer(evtID)); }

      if(evtInfo!=null)
      {
         synchronized( idToEvtInfo ) { idToEvtInfo.remove(new Integer(evtID)); }
         synchronized( evts ) { evts.removeElement(evtInfo); }
         numberOfEvents--;

      }// end of if

   }// end of cancelEvent()

   /**
   */
   final synchronized void
   log(String label, RTIEvent evt)
   {
      String logFileName = new String("/tmp/fed"+fedInf.getFederateID()+".log");
      try
      {
        PrintWriter pw = new PrintWriter(new FileWriter(logFileName,true),true);
        pw.println(label+" "+evt.toString());
        pw.close();

      } catch(java.io.IOException ioe) {}

   }// end of logEvent()
 
   /**
   */
   public void
   setAsynchronousDelivery()
   {
       aDelivery = true;
       start();

   }// end of setAsynchronousDelivery()

   /**
   */
   public void
   resetAsynchronousDelivery()
   {
       aDelivery = false;
       stop();

   }// end of resetAsynchronousDelivery()

   // Runnable Interface
   public void
   start()
   { 
      if(runner==null) runner = new Thread(this);
      run_stay = true;
      runner.start();

   }// end of start()

   public void
   stop()
   {
      run_stay = false;
      if(runner!=null) runner.stop(); 
      runner = null; 

   }// end of stop()

   public void
   run()
   {

     while( run_stay )
     {
        while( idToEvtInfo.isEmpty() ) 
        {
           try
           { runner.sleep(100); }  //wait()
           catch(InterruptedException ie) {}
        } 
           try
           { runner.sleep(300); }  //wait()
           catch(InterruptedException ie) {}

        if( !idToEvtInfo.isEmpty() ) processEvents(fedAmb); 

     }// end of while

   }// end of run()

/*
     if( !idToEvtInfo.isEmpty() ) 
     {
        Vector    evts0 = new Vector();
 
        Hashtable idToEvtInfo0 = new Hashtable();

        synchronized( evts ) 
        { evts0=evts; evts = new Vector();}

        synchronized( idToEvtInfo ) 
        { idToEvtInfo0=idToEvtInfo; idToEvtInfo = new Hashtable();}

        processEvents(evts0,idToEvtInfo0,fedAmb);

     }// end of if

     stop();

   }// end of run()
*/

   /**
   * Executes  all the events in the EventQue
   */
   final void 
   processEvents( Vector evts0, Hashtable idToEvtInfo0, 
                  RTIcap.FederateAmbassador _fed)
   {

      int   num_of_exec = 0;
      EventInfo evtInf;
      RTIEvent evtInstance;
      Enumeration e = evts0.elements();;

      if(!e.hasMoreElements() && DEBUG) 
        System.out.println("There is nothing to process for "+FederateName+"!");

      for(;e.hasMoreElements();) //&&num_of_exec<10;)
      {
          evtInf = (EventInfo) e.nextElement();
          if(evtInf!=null) 
          { 
            evtInstance = null;

            /** Find the RTIEvent object */
            if(que!=null) evtInstance = que.getRTIEvent(evtInf.getHandle());

            if(evtInstance!=null) 
            {
               log(" Execute ",evtInstance);

               // Note: We need a way to inform that this federate is no
               // longer in the federation execution. Question is
               // can we conclude that this federate is not responsive 
               // from the first exception or should we try to do more
               // fancy things. We can anounce that this federate
               // left the building as soon as we catch the first exception
               // about federateAmbassador when it is not responsing. 
               /** Execute this event */
               evtInstance.execute(_fed);
               num_of_exec++;
 
               /** Delete this record from the Event List */
               //synchronized( evts ) { evts.removeElement(evtInf); }
               evts0.removeElement(evtInf); 

               /** Delete this record from the Event List */
               //synchronized( idToEvtInfo )
               //{ idToEvtInfo.remove(new Integer(evtInf.getHandle())); }
               idToEvtInfo0.remove(new Integer(evtInf.getHandle())); 

               //numberOfEvents--;
 
            }// end of if 

            evtInf = null;

          }// end of if

      }// end of for

      e = null;

      //return num_of_exec;

   }// end of processEvents()

}// end of class EvtQueueRO

