package owrti;
import owrti.evt.*;
import RTIcap.*;
import java.util.*;
import java.io.*;

/**
*
*  For each interaction there is one Routing Space definition for it.
*    
*  Maybe we should define a default region in each routing space such
*  this region covers the whole space.
* 
* @date April 15, 1998 ( Updated at May 02, 1998)
* @author H. Timucin Ozdemir
*
*/
public class InteractionRecord extends RegionSubscriberRegistry implements RTIDebug
{
               short  interactionId;
              String  interactionClassName;
     ParameterNaming  paramsDirectory;   
   InteractionRecord  father;

   /** Regions defined in this Routing Space Q: Do I really want to keep it? */
   Hashtable regions = new Hashtable();

   RTIMessageDispatcher dispatcher;

   /** */
   public InteractionRecord() { this(new SpaceInfo()); }

   /** */
   public InteractionRecord(SpaceInfo _space) {
      super(_space);
      paramsDirectory = new ParameterNaming(this);
      dispatcher=new RTIMsgDispatcher();
   }// end of CONSTRUCTOR

   /** */
   public String getName() { return interactionClassName; }

   /** */
   public short getId() { return interactionId; }

   /** */
   public void setName(String name) { interactionClassName=name; }

   /** */
   public void setId(short _id) { interactionId=_id; }

    /** */
    public short addParameter(String theName) {
       return paramsDirectory.addParameter(theName);
    }

    /** */
    public short getParameterHandle(String theName)
          throws RTIcap.NameNotFound, RTIcap.FederateNotExecutionMember, 
                 RTIcap.RTIinternalError {
       return paramsDirectory.getParameterHandle(theName);
    }

    /** */
    public String 
    getParameterName(short theHandle)
           throws RTIcap.InteractionParameterNotDefined, 
                  RTIcap.FederateNotExecutionMember, RTIcap.RTIinternalError 
    { return paramsDirectory.getParameterName(theHandle); }

    /** 
    * Broadcast this message to all queues 
    */
    public void
    publishParams( RTIcap.ParameterHandleValuePair[] theParameters, 
                   double theTime, String theTag, FederateInfo _fedHandle) {
      publishParamsWithRegion( theParameters, theTime, theTag, null,_fedHandle) ;
    }// end of publishParams()

    /*********************************/
    /** Data Distribution Management */
    /*********************************/

    /**
    *
    * 1. Add this region to the subscribing regions list if it does not
    *    exist.
    *
    * 2. Add Federate to the subscriber list of this region.
    *
    */
    public void associateWithRegion(RegionInfo regInf, FederateInfo fedInf) {
       if( regions.get(new Integer(regInf.getHandle())) == null) 
           regions.put(new Integer(regInf.getHandle()),regInf);
       regInf.register_subscriber(fedInf);
    }// end of associateWithRegion() 

    /**
    *
    * 1. If subscribing region exist,  then remove this federate from 
    *    Region's subscription list.
    * 
    * Q: What happens if region has 0(ZERO) subscribers? Should I remove it?
    * 
    */
    public void unassociateFromRegion(RegionInfo regInf, FederateInfo fedInf) {
       if( regions.get(new Integer(regInf.getHandle())) != null) 
           regInf.deregister_subscriber(fedInf.getFederateID());
    }// end of unassociateFromRegion() 

    final void copy(Enumeration e, Vector feds) {
       Object o;
       if(e!=null)
       for(;e.hasMoreElements();) { 
          o=e.nextElement();
          if(o!=null&&!feds.contains(o)) feds.addElement(o);
       }// end of for
    }// end of copy()

    /** 
    * Broadcast this message to all queues in the intersecting regions. 
    *
    * Ask SpaceInfo about the possible subscribers which overlap with the
    * region.
    *
    */
    public void
    publishParamsWithRegion( RTIcap.ParameterHandleValuePair[] theParameters, 
                             double theTime, String theTag, RegionInfo regInf,
                             FederateInfo _fedHandle) {
       if(DEBUG) System.out.println(" In publishParamsWithRegion() ");
       Vector feds = new Vector();
       // ** Send everybody in the defeault region * 
       copy(getSubscribers(),feds);
       Enumeration e = regions.elements();
       RegionInfo rgInf ;
       if( e!= null)
         for(;e.hasMoreElements();) {
            rgInf = (RegionInfo) e.nextElement();
            // **************************************************************
            // ** If there is intersection with the subscription region and *
            // ** the publishing region then we have to to inform all       *
            // ** tenants of the subscription region.                       *
            // **************************************************************
            if(rgInf!=null&&(regInf==null||regInf.canIntersect(rgInf))) 
               copy(rgInf.getSubscribers(),feds);
         }// end of for
       dispatcher.addRTIMsg(new IRPublishParameters(theParameters, theTime, 
                                   theTag,  _fedHandle, feds.elements(),
                                   _fedHandle.getFederateID()) );
       if(DEBUG) System.out.println(" Out of publishParamsWithRegion() ");
    }// end of publishParamsWithRegion()

    /**   */
    public void dump(OutputStream out) {
      System.out.println(" Interaction Name : "+interactionClassName);
      System.out.println(" Interaction Id   : "+interactionId);
      int numOfParams = paramsDirectory.getNumberOfParameters(); 
      System.out.println(" # of Parameters : "+numOfParams);
      for(int i=1;i<(numOfParams+1);i++)
      try {
         System.out.println(i+"th params name : "+paramsDirectory.getParameterName((short)i));
      } catch(RTIcap.InteractionParameterNotDefined fnem) {}
    }// end of dump()

   class IRPublishParameters implements RTIMessage, RTIDebug {
     RTIcap.ParameterHandleValuePair[] theParameters; 
     double theTime; String theTag; FederateInfo _fedHandle;
     Enumeration e; short except;
     IRPublishParameters(RTIcap.ParameterHandleValuePair[] theParameters, 
                   double theTime, String theTag, FederateInfo _fedHandle,
                   Enumeration e, short except) {
       this.theParameters=theParameters;this.theTime=theTime;
       this._fedHandle=_fedHandle;this.e=e;this.except=except;
     }
     public void execute() {
       RTIcap.EventRetractionHandle retract = new RTIcap.EventRetractionHandle(theTime,0,_fedHandle.getFederateID());
       FederateInfo n ;
       if(e!=null)
         for(;e.hasMoreElements();) {
            n = (FederateInfo) e.nextElement();
            if(n!=null && n.getFederateID()!=except) {
               n.addRTIEvent( new EvtReceiveInteraction(interactionId,theParameters,theTag,retract,theTime));
               if(DEBUG) System.out.println(" I sent to "+n.getFederateID());
            }// end of if
         }// end of for
     }// end of execute()
   }// end of IRPublishParameters

}// end of class InteractionRecord


