package owrti;

import RTIcap.*;
import java.util.*;

/**
*
* Keeps the ObjectClass-Instance to ObjectRecord mapping.
*
* @date April 29, 1998 ( Updated at May 02, 1998)
* @author H. Timucin Ozdemir
*
*/
public class 
ObjectClassNaming implements RTIDebug
{

   /** Class Id -> Class Name */
   Hashtable idToName = new Hashtable();

   /** Class Name -> Class Id */
   Hashtable nameToId = new Hashtable();

   /** Class Id -> Object Class Record */
   Hashtable idToObjRecord = new Hashtable();
   Object    objRecordLock = new Object();

   /** Instance Id -> Object Class Id */
   Hashtable objInsIdToClassId = new Hashtable();

   /** Instance Name -> Instance Id */
   Hashtable objNameToObjInsId = new Hashtable();
   Object    objInstanceLock = new Object();

   InteractionNaming interNaming = null;   /// ?????

   /**
   */
   public 
   ObjectClassNaming(InteractionNaming _interNaming) 
   { interNaming = _interNaming; }

   /**
   * returns the Object Id of the particular Object Name
   */
   public short 
   getObjectClassHandle(String theName) throws RTIcap.NameNotFound
   {
       /**
       */

       String lookUpName = theName;
       Short ret=null;

       if(theName.startsWith("objectroot")||theName.startsWith("objectRoot")) 
       {
          lookUpName = "objectRoot."+theName.substring(11,theName.length());
          ret=(Short) nameToId.get(lookUpName); 
          System.out.println(" I asked "+theName+" and searched "+lookUpName);
       }
       else 
       {
          lookUpName = "objectRoot."+theName;
          ret = (Short) nameToId.get(lookUpName); 
          System.out.println(" I asked "+theName+" and searched "+lookUpName);
       }

       if(ret==null) throw new RTIcap.NameNotFound();
       System.out.println(" The handle of "+lookUpName+" is "+ret.shortValue());

       return ret.shortValue();
    }

   /**
   */
   public String 
   getObjectClassName(short theHandle) throws RTIcap.ObjectClassNotDefined 
   {
       String ret = (String) idToName.get(new Short(theHandle));

       if(ret==null) throw new RTIcap.ObjectClassNotDefined();

       return ret;

   }// end of getObjectClassName() 

   /**
   */
   public short 
   getAttributeHandle(String theName, short whichClass)
          throws RTIcap.ObjectClassNotDefined, RTIcap.NameNotFound 
   {
       ObjectRecord ret = (ObjectRecord) idToObjRecord.get(new Short(whichClass));

       if(ret==null) throw new RTIcap.ObjectClassNotDefined();

       return ret.getAttributeHandle(theName);

   }// end of getAttributeHandle()

   /**
   */
   public String 
   getAttributeName(short theHandle, short whichClass)
           throws RTIcap.ObjectClassNotDefined, RTIcap.AttributeNotDefined 
   {
       ObjectRecord ret = (ObjectRecord) idToObjRecord.get(new Short(whichClass));

       if(ret==null) throw new RTIcap.ObjectClassNotDefined();

       return ret.getAttributeName(theHandle);

   }// end of getAttributeName()

   /**
   */
   public final ObjectRecord 
   getObjectRecord( short whichClass ) throws RTIcap.ObjectClassNotDefined {
       ObjectRecord ret = (ObjectRecord) idToObjRecord.get(new Short(whichClass));
       if(ret==null) throw new RTIcap.ObjectClassNotDefined();
       return ret;
   }// end of getObjectRecord()

   /**
   */
   public final ObjectRecord 
   getObjectRecord( String objectClassName ) throws RTIcap.ObjectClassNotDefined {
      Short classId = (Short) nameToId.get(objectClassName);
      if( classId==null ) throw new RTIcap.ObjectClassNotDefined();
      return getObjectRecord(classId.shortValue());
   }// end of getObjectRecord()

   /**
   */
   public final void addObjectRecord(ObjectRecord rec) {
      synchronized(objRecordLock) {
        idToName.put(new Short(rec.getId()),rec.getName());
        nameToId.put(rec.getName(),new Short(rec.getId()));
        idToObjRecord.put(new Short(rec.getId()),rec);
      }// end of synchronized()
   }// end of addObjectRecord()

   /**
   */
   public final void removeObjectRecord(ObjectRecord rec) {
      synchronized(objRecordLock) {
        idToName.remove(new Short(rec.getId()));
        nameToId.remove(rec.getName());
        idToObjRecord.remove(new Short(rec.getId()));
      }// end of synchronized()
   }// end of removeObjectRecord()

   /**
   */
   public final void removeObjectRecord(short recId) {
      try 
      { removeObjectRecord(getObjectRecord(recId));
      } catch(RTIcap.ObjectClassNotDefined ocnd) {}
   }// end of removeObjectRecord()

   /**
   */
   public final void removeObjectRecord( String objectClassName ) {
      try 
      { removeObjectRecord(getObjectRecord(objectClassName));
      } catch(RTIcap.ObjectClassNotDefined ocnd) {}
   }// end of removeObjectRecord()

   /**
   */
   public final void addObjectInstance(ObjectInstanceRecord rec) {
      synchronized(objInstanceLock) {
        objInsIdToClassId.put( new Integer(rec.getObjectInstanceHandle()), 
                               new Short(rec.getObjectClassHandle()) );
        objNameToObjInsId.put( rec.getObjectName(), new Integer(rec.getObjectInstanceHandle())); 
      }// end of synchronized()
   }// end of addObjectInstance()

   /**
   */
   public final void removeObjectInstance(ObjectInstanceRecord rec) {
      synchronized(objInstanceLock) {
        objInsIdToClassId.remove( new Integer(rec.getObjectInstanceHandle())); 
        objNameToObjInsId.remove( rec.getObjectName() ); 
      }// end of synchronized()
   }// end of removeObjectInstance()

   /**
   */
   public final void 
   removeObjectInstance( int theHandle ) throws RTIcap.ObjectNotKnown {
      ObjectInstanceRecord rec = getObjectInstance(theHandle);
      if( rec==null ) throw new  RTIcap.ObjectNotKnown();
      removeObjectInstance(rec);
   }// end of removeObjectInstance()

   /**
   */
   public final void 
   removeObjectInstance( String theName ) throws RTIcap.ObjectNotKnown {
      ObjectInstanceRecord rec = getObjectInstance(theName);
      if( rec==null ) throw new  RTIcap.ObjectNotKnown();
      removeObjectInstance(rec);
   }// end of removeObjectInstance()

   /**
   public final ObjectInstanceRecord getObjectInstance(int theHandle) {
      ObjectInstanceRecord ret = null; 
      ObjectRecord rec = (ObjectRecord) idToObjRecord.get(new Short((short)theHandle)); 
      if(rec!=null) ret = rec.getInstance(theHandle); 
      return ret; 
   }// end of addObjectInstance()
   */

   /**
   *  Find its class id
   *  Finds its ObjectRecord
   *  ask ObjectRecord to provide the particular ObjectInstance
   */
   public final ObjectInstanceRecord 
   getObjectInstance(int theHandle) throws RTIcap.ObjectNotKnown {
      ObjectInstanceRecord ret = null; 
      Short ClassId = (Short) objInsIdToClassId.get( new Integer(theHandle) );
      if( ClassId == null ) throw new RTIcap.ObjectNotKnown();
      ObjectRecord rec = (ObjectRecord) idToObjRecord.get(new Short(ClassId.shortValue())); 
      if( rec == null ) throw new RTIcap.ObjectNotKnown();
      System.out.println("ObjectRecord found for "+ClassId.shortValue());
      ret = rec.getInstance(theHandle); 
      if( ret == null ) throw new RTIcap.ObjectNotKnown();
      System.out.println("ObjectInstanceRecord found for "+theHandle);
      return ret; 
   }// end of getObjectInstance()

   /**
   *  Find its class id
   *  Finds its ObjectRecord
   *  ask ObjectRecord to provide the particular ObjectInstance
   */
   public final ObjectInstanceRecord 
   getObjectInstance( String theName ) throws RTIcap.ObjectNotKnown {
      Integer objId = (Integer) objNameToObjInsId.get(theName);
      if( objId == null ) throw new RTIcap.ObjectNotKnown();
      return getObjectInstance(objId.intValue()); 
   }// end of getObjectInstance()

}// end of class
