/**
* Keeps the instance number, object class number, federate handle who
* defined this instance.
*/
package owrti;
import owrti.evt.*;
import RTIcap.*;
import java.util.*;

/**
* 
* @date April 20, 1998 ( Last Updated March 31, 1999)
* @author H. Timucin Ozdemir
*
*/
public class ObjectInstanceRecord extends ObjectRecord implements RTIDebug {

   short  privelegeToDelete = -1;

   /** One who defined this object at the first place */
   short  theFedHandle;

   /** Instance of Which Object Class */
   short  theObjectClassHandle;

   /** Object Instance Id */
   int    theInstanceHandle;

   /** Object Instance Name */
   String    theInstanceName;

   FederateInfo fedInfo;

   ObjectRecord objFactory;

   /** Attribute Information center */
   AttributeNaming attributeDirectory;

   AttributeInfoNaming myAttributes ;

   /** */
   Hashtable attWatcher = new Hashtable();

   /** */
   RegionInfo myRegion;

   RTIMessageDispatcher dispatcher;

   /**
   */
   public ObjectInstanceRecord(short _theFedHandle, short _theObjectClassHandle,
                        int _theInstanceHandle, FederateInfo _father,
                        ObjectRecord _objFactory, RTIMessageDispatcher _dispatcher,
                        RegionInfo _myRegion ) {
      this( _theFedHandle, _theObjectClassHandle, _theInstanceHandle, _father, 
            _objFactory, _dispatcher,_myRegion, 
            new String( _objFactory.getName()+_theInstanceHandle ) );
   }// end of CONSTRUCTOR

   /**
   */
   public ObjectInstanceRecord(short _theFedHandle, short _theObjectClassHandle,
                         int _theInstanceHandle, FederateInfo _father,
                         ObjectRecord _objFactory, RTIMessageDispatcher _dispatcher,
                         RegionInfo _myRegion, String theName ) {
      myAttributes = new AttributeInfoNaming(this);
      theFedHandle         = _theFedHandle;
      theObjectClassHandle = _theObjectClassHandle;
      theInstanceHandle    = _theInstanceHandle;
      fedInfo               = _father;
      privelegeToDelete    = theFedHandle;
      objFactory           = _objFactory;
      myRegion             = _myRegion;
      dispatcher           = _dispatcher;
      try
      {  attributeDirectory   = objFactory.getAttributeNaming((short)0);
      } catch(RTIcap.RTIinternalError fnem) {}
      defineAttributeWatchers();
      theInstanceName = new String( theName );
   }// end of defineObjectInstance()

   public RegionInfo getRegionInfo() { return myRegion; }

   /** */
   final void defineAttributeWatchers() {
      if(attributeDirectory==null) return;
      Enumeration e = attributeDirectory.getAllNames();
      short ret ;
      String name; 
      for(;e.hasMoreElements();) {
         name = (String) e.nextElement();
         if(name!=null) {
            try
            {  ret = attributeDirectory.getAttributeHandle(name);
               if(ret>=0) {
                  attWatcher.put(new Short(ret),new AttributeWatcher(this,ret,name,theFedHandle));
                  myAttributes.addAttributeInfo(new AttributeInfo(this,ret,name,
                               theFedHandle,
                               objFactory.getDefaultRegionForAttribute(ret),
                               RTIcap.TransportType.BEST_EFFORT,
                               RTIcap.OrderType.RECEIVE));
               } // end of if
            } catch(RTIcap.NameNotFound fnem) {}
         }// end of if
      }// end of for
   }// end of defineAttributeWatchers()

   /** */
   public String getObjectName() { return theInstanceName; }

   /** */
   public short getFederateHandle() { return theFedHandle; }

   /**
   */
   public FederateAmbassador 
   getFederateAmbassador() { return fedInfo.getFederateAmbassador(); }

   /** */
   public FederateInfo getFederateInfo() { return fedInfo; }

   /** */
   public short getObjectClassHandle() { return theObjectClassHandle; }

   /** */
   public int   getObjectInstanceHandle() { return theInstanceHandle; }

   /** */
   public short   getOwnerHandle() { return privelegeToDelete; }

   /** */
   public ObjectRecord   getObjectClass() { return objFactory; }

   /**
   */
   public final AttributeInfo   
   getAttributeInfo(short attributeHandle ) 
   { return myAttributes.getAttributeInfo(attributeHandle); }

   /**
   */
   public void  releaseOwnerships(short[] attributes, boolean unconditional) { 
       AttributeInfo attInfele;

       for(int i=0;i<attributes.length;i++) {
          attInfele = getAttributeInfo(attributes[i]); 
          if(attInfele!=null) {   
             if(unconditional) { 
                attInfele.setOwnerId((short)-1); attInfele.resetOwned();
             }// end of if
             attInfele.setOwnershipTransferPermission();
          }// end of if
       }// end of for

   }// end of releaseOwnerships()

   /**
   */ 
   public final AttributeWatcher   
   getAttributeWatcher(short attribIndex)
   { return (AttributeWatcher) attWatcher.get(new Short(attribIndex)); }

   /**
   *
   * Go through all the attributeInfo objects in the attributeinfonaming
   * and set their region definition to the default region definition if
   * this attribute is in the given region.
   *
   */ 
   public void   unassociateAttributesFromRegion(RegionInfo theRegion) { 
          Enumeration   e = attWatcher.keys();
                Short   key;
        AttributeInfo   att;
      AttributeWatcher  aWatch;

      for(;e.hasMoreElements();) {
         key = (Short) e.nextElement();
         if(key!=null) {
            aWatch = (AttributeWatcher) attWatcher.get(key);
            att = (AttributeInfo) myAttributes.getAttributeInfo(key.shortValue());
            if( att != null && aWatch != null ) 
                att.setRegionInfo(aWatch.getDefaultRegion());
         }// end of if 
      }// end of for
 
   }// end of unassociateAttributesFromRegion()

   /**
   * 
   */ 
   public void   
   associateAttributesWithRegions(short[] attributes, RTIcap.Region rgn, 
          SpaceNaming spaceNaming ) { 

       AttributeInfo crr;
       RegionInfo    regInf = spaceNaming.getRegionInfo(rgn);
       if( regInf != null )
         for(int i=0;i<attributes.length;i++) {
            crr = myAttributes.getAttributeInfo(attributes[i]);
            if( crr != null ) crr.setRegionInfo(regInf);
         }// end of for

   }// end of associateAttributesWithRegions()

   /**
   */ 
   public void 
   associateAttributesWithRegions(short[] attributes, RTIcap.Region[] rgns, 
          SpaceNaming spaceNaming )
   {  AttributeInfo crr;
      RegionInfo regInf ;
      for(int i=0; (i<attributes.length) && (i<rgns.length) ;i++) {
         crr = myAttributes.getAttributeInfo(attributes[i]);
         regInf = spaceNaming.getRegionInfo(attributes[i]);
         if(regInf!=null && crr!=null) crr.setRegionInfo(regInf);
      }// end of for
   }// end of unassociateFromRegion()

   /**
   *  For each attribute
   *  1. Get its Owner. If nobody owns it, then exit.
   *  2. Get its Region
   *  3. If its region is null or intersects with the given region.
   *     then send 'provideAtt..' message to this federate.
   */
   public void
   requestUpdateAttributesInRegion(short[] attributes, RegionInfo regInf) {
      dispatcher.addRTIMsg(new OIRRequestUpdateAttributeInRegion(attributes,regInf));
   }// end of requestUpdateAttributesInRegion() 

   /**
   *  For each attribute, ask its owner to provide update for that
   *  particular attribute.
   *  If attribute has not an owner, then there is nobody to ask.
   */
   public void
   requestObjectAttributeValueUpdate(short[] theAttributes) {
      requestUpdateAttributesInRegion2(theAttributes, null);
   }// end of requestObjectAttributeValueUpdate()

   /**
   *  For each attribute
   *  1. Get its Owner. If nobody owns it, then exit.
   *  2. Get its Region
   *  3. If its region is null or intersects with the given region.
   *     then send 'provideAtt..' message to this federate.
   */
   public void
   requestUpdateAttributesInRegion2(short[] attributes, RegionInfo regInf) {
     dispatcher.addRTIMsg(new OIRRequestUpdateAttributeInRegion(attributes,regInf));
   }// end of requestUpdateAttributesInRegion2() 

   /**
   *  For each attribute
   *  1. Get its Owner. If nobody owns it, then exit.
   *  2. Get its Region
   *  3. If its region is null or intersects with the given region.
   *     then send 'provideAtt..' message to this federate.
   */
   public void
   turnUpdatesOnOff(short[] attributes, boolean OnOff, RegionInfo regInf) {
     dispatcher.addRTIMsg(new OIRTurnUpdatesOnOff(attributes,OnOff,regInf));
   }// end of turnUpdatesOnOff() 

   /**
   * Returns the Federate Id of Owner of given attribute
   */
   public short getAttributeOwner( short theHandle ) {
      AttributeInfo att = getAttributeInfo(theHandle); 
      return ( att!=null? att.getOwnerId() : -1 );
   }// end of getAttributeOwner()

   /** */
   public boolean isItOwner( short theHandle , short fedHandle ) {
      AttributeInfo att = getAttributeInfo(theHandle); 
      boolean ret = ( att!=null );
      return (ret?(att.getOwnerId()==fedHandle):ret);
   }// end of isItOwner()

   /**
   * Returns the Federate Id of Owner of given attribute
   */
   public short[] getAllAttributeHandles( ) {
      return objFactory.getAllAttributeHandles();
   }// end of getAllAttributeHandles()

   class OIRGenericMessage implements RTIMessage {
     short[] attributes; RegionInfo regInf; short type;
     OIRGenericMessage(short[] attributes, RegionInfo regInf, short type) {
       this.attributes=attributes; this.regInf=regInf;this.type=type;
     }
     public void execute() {
       AttributeInfo    att;
       RegionInfo       r;
       short            fid;
       FederateInfo     fdInf;
       FederationExecution fedEx = fedInfo.getFederationExecution();
       short[]          val = new short[1];
       if(type>=0&&type<3)
       for(int i=0;i<attributes.length;i++) {
         att = myAttributes.getAttributeInfo(attributes[i]);
         if(att!=null) {
            fid = att.getOwnerId();
            if(fid>=0) {
               fdInf = fedEx.getFederateInfo(fid);
               if(fdInf!=null) {
                  r = att.getRegionInfo();
                  val[0] = attributes[i];
                  if( r == null || regInf==null || r.canIntersect(regInf) )
                     switch(type) {
                      case 0:  fdInf.addRTIEvent(new EvtTurnUpdatesOnForObjectInstance(theInstanceHandle,val));
                               break;
                      case 1:  fdInf.addRTIEvent(new EvtTurnUpdatesOffForObjectInstance(theInstanceHandle,val));
                               break;
                      case 2:  fdInf.addRTIEvent(new EvtProvideAttributeValueUpdate(theInstanceHandle,val));
                               break;
                     }// end of switch
               }// end of if
            }// end of if
         }// end of if
       }// end of for
     }// end of execute()
   }// end of class OIRGenericMessage

   class OIRRequestUpdateAttributeInRegion extends OIRGenericMessage {
     OIRRequestUpdateAttributeInRegion(short[] attributes, RegionInfo regInf) {
       super(attributes,regInf,(short)2);
     }// end of CONSTRUCTOR
   }// end of class OIRRequestUpdateAttributeInRegion

   class OIRTurnUpdatesOnOff extends OIRGenericMessage {
     OIRTurnUpdatesOnOff(short[] attributes, boolean OnOff, RegionInfo regInf) {
       super(attributes,regInf,(short)(OnOff?0:1)); 
     }// end of switch
   }// end of class OIRTurnUpdatesOnOff

}// end of ObjectInstanceRecord
