/**
*
*  RTI 1.3 Object Manager
*  May 1st, to make it RTI 1.3 Complaint
*/
package owrti;
import RTIcap.*;
import npacrti.*;
import java.util.*;
/*
* @date April 17, 1998 ( Last Updated March 31, 1999)
* @author H. Timucin Ozdemir
*/
public class 
ObjectManager extends _ObjectManagerInterfaceImplBase implements RTIDebug {

    FederationExecution fedEx;
    ObjectClassNaming objNaming;
    InteractionNaming interActionNaming;
    int MAX_OBJECTS_PER_FEDERATE = 1000;
    int numberOfObjects = 1;
    int eventCount = 0;
    Hashtable objIdToClassId = new Hashtable();

    /**
    */
    public 
    ObjectManager( ObjectClassNaming _objNaming, 
                   InteractionNaming _interActionNaming,
                   FederationExecution _fedEx)
    {
        objNaming = _objNaming;
        interActionNaming = _interActionNaming;
        fedEx = _fedEx;

    }// end of CONSTRUCTOR

    /**
    * Flush Queues
    */
    public void 
    flushEvents() 
    {
    }// end of flushEvents() 

    /**
    * 
    */
    public
    void requestID(short idCount, org.omg.CORBA.IntHolder firstID, 
                   org.omg.CORBA.IntHolder lastID)
         throws RTIcap.TooManyIDsRequested, RTIcap.IDsupplyExhausted, 
                RTIcap.FederateNotExecutionMember, 
                RTIcap.ConcurrentAccessAttempted, RTIcap.RTIinternalError, 
                RTIcap.Deadlock
    {
       firstID.value = numberOfObjects;
       numberOfObjects += idCount-1;
       lastID.value = numberOfObjects;
       numberOfObjects++;
       if(DEBUG)
       {
          System.out.println(" firstID "+firstID.value);
          System.out.println(" lastID "+lastID.value);
       }

    }// end of requestID()

    /**
    * 6.2
    */
    public synchronized
    void registerObject(short theClass, int theObject,
                FederateAmbassador _fedHandle, short _fedId )
         throws RTIcap.ObjectClassNotDefined, RTIcap.ObjectClassNotPublished, 
                RTIcap.InvalidObjectID, RTIcap.ObjectAlreadyRegistered, 
                RTIcap.FederateNotExecutionMember, 
                RTIcap.ConcurrentAccessAttempted, RTIcap.RTIinternalError, 
                RTIcap.Deadlock
    {

       if(DEBUG) System.out.println(" in registerObject() for class "+theClass+" Object "+theObject);
       isExecutionMember(_fedId); 

       /** Find the Record of this object */
       ObjectRecord rec = objNaming.getObjectRecord(theClass); 

       /** Register this object to this Class */
       rec.register_object(theObject,null,fedEx.getFederateInfo(_fedId),_fedId,fedEx);
       ObjectInstanceRecord aRec = rec.getInstance(theObject);
       if( aRec!=null ) objNaming.addObjectInstance(aRec);

       /** Keeps the mapping between ObjectID and ClassID */ 
       objIdToClassId.put(new Integer(theObject),new Short(theClass));

       if(DEBUG) System.out.println(" out of registerObject() ");

    }// end of registerObject()

    /**
    * Propriatery
    */ 
    synchronized int 
    registerObject(short theClass, String theName, short _fedId )
         throws RTIcap.ObjectClassNotDefined, RTIcap.ObjectClassNotPublished, 
                RTIcap.ObjectAlreadyRegistered, RTIcap.RTIinternalError,
                RTIcap.FederateNotExecutionMember
    {
       int theObject = numberOfObjects;

       if(DEBUG) System.out.println(" in registerObject() for class "+theClass+" Object "+theObject);
       isExecutionMember(_fedId); 

       /** Find the Record of this object */
       ObjectRecord rec = objNaming.getObjectRecord(theClass); 

       try{ if(theName!=null&&objNaming.getObjectInstance(theName)!=null)
               throw new RTIcap.ObjectAlreadyRegistered();
       } catch(RTIcap.ObjectNotKnown onk) {}

       if(!rec.canPublish(_fedId)) throw new RTIcap.ObjectClassNotPublished();

       /** Register this object to this Class */
       rec.register_object(theObject,theName,fedEx.getFederateInfo(_fedId),_fedId,fedEx);
       ObjectInstanceRecord aRec = rec.getInstance(theObject);
       if( aRec!=null ) objNaming.addObjectInstance(aRec);

       /** Keeps the mapping between ObjectID and ClassID */ 
       objIdToClassId.put(new Integer(theObject),new Short(theClass));

       numberOfObjects++;

       if(DEBUG) System.out.println(" out of registerObject() ");

       return theObject;

    }// end of registerObject()

    /**
    *  RTI 1.3
    */
    public int 
    registerObjectInstanceWithName(short theClass, String theName,
           RTIcap.FederateAmbassador fedAmb, short theHandle)
       throws RTIcap.ObjectClassNotDefined, RTIcap.ObjectClassNotPublished, 
              RTIcap.ObjectAlreadyRegistered, RTIcap.FederateNotExecutionMember,
              RTIcap.SaveInProgress, RTIcap.RestoreInProgress, 
              RTIcap.RTIinternalError
    {
        return registerObject(theClass,theName,theHandle);

    }// end of registerObjectInstanceWithName()

    /**
    *  RTI 1.3
    */
    public int 
    registerObjectInstance(short theClass, RTIcap.FederateAmbassador fedAmb , 
           short theHandle)
        throws RTIcap.ObjectClassNotDefined, RTIcap.ObjectClassNotPublished, 
               RTIcap.FederateNotExecutionMember, RTIcap.SaveInProgress,
               RTIcap.RestoreInProgress, RTIcap.RTIinternalError {
        int ret=-1;
        try { ret = registerObject(theClass,null,theHandle); } 
        catch(RTIcap.ObjectAlreadyRegistered oar) 
        { throw new RTIcap.RTIinternalError(); }
        return ret;
    }// end of registerObjectInstance()

    /**
    * 6.4
    */
    RTIcap.EventRetractionHandle 
    updateAttributeValues(int theObject, 
                RTIcap.AttributeHandleValuePair[] theAttributes, 
                double theTime, String theTag,
                FederateAmbassador _fedHandle, short _fedId )
         throws RTIcap.ObjectNotKnown, RTIcap.AttributeNotDefined, 
                RTIcap.AttributeNotOwned, RTIcap.InvalidFederationTime, 
                RTIcap.FederateNotExecutionMember, 
                RTIcap.ConcurrentAccessAttempted, RTIcap.RTIinternalError, 
                RTIcap.Deadlock
    {  int i;
       if(DEBUG) 
          System.out.println(" in updateAttributeValues() of ObjectManager for "+theObject);
       isExecutionMember(_fedId); 
       RTIcap.EventRetractionHandle ret = null;
       Short theClassId = (Short) objIdToClassId.get(new Integer(theObject));

       if( theClassId==null ) throw new RTIcap.ObjectNotKnown();
       try
       { ObjectRecord rec = objNaming.getObjectRecord(theClassId.shortValue());
         ObjectInstanceRecord oRec = objNaming.getObjectInstance(theObject);
         for(i=0;i<theAttributes.length;i++) 
             rec.haveAttribute(theAttributes[i].handle);
         for(i=0;i<theAttributes.length;i++) 
           if(!oRec.isItOwner(theAttributes[i].handle,_fedId))
               throw new RTIcap.AttributeNotOwned();
         eventCount++;
         ret = new EventRetractionHandle(theTime,eventCount,_fedId);
         rec.updateAttributeValues(theObject,theAttributes,theTime,theTag,_fedHandle,_fedId);
       } catch(RTIcap.ObjectClassNotDefined ocnd) 
        { throw new RTIcap.ObjectNotKnown();}

       if(DEBUG) System.out.println(" out of updateAttributeValues() of ObjectManager ");
       return ret;
    }// end of updateAttributeValues()

    /**
    * RTI 1.3
    */
    public RTIcap.EventRetractionHandle 
    updateAttributeValuesWithTime(int theObject, 
           RTIcap.AttributeHandleValuePair[] theAttributes, double theTime, 
           String theTag, RTIcap.FederateAmbassador fedAmb, short _fedId)
        throws RTIcap.ObjectNotKnown, RTIcap.AttributeNotDefined,
               RTIcap.AttributeNotOwned, RTIcap.InvalidFederationTime,
               RTIcap.FederateNotExecutionMember, RTIcap.SaveInProgress, 
               RTIcap.RestoreInProgress, RTIcap.RTIinternalError
    {  int i;
       if(DEBUG) 
          System.out.println(" in updateAttributeValuesWithTime() of ObjectManager for "+theObject);

       isExecutionMember(_fedId); 
       RTIcap.EventRetractionHandle ret = null;
       Short theClassId = (Short) objIdToClassId.get(new Integer(theObject));
       if( theClassId==null ) throw new RTIcap.ObjectNotKnown();
       try
       { ObjectRecord rec = objNaming.getObjectRecord(theClassId.shortValue());
         ObjectInstanceRecord oRec = objNaming.getObjectInstance(theObject);
         for(i=0;i<theAttributes.length;i++) 
             rec.haveAttribute(theAttributes[i].handle);
         // TIMUCIN 05.04.1999
         //for(i=0;i<theAttributes.length;i++) 
         //  if(!oRec.isItOwner(theAttributes[i].handle,_fedId))
         //      throw new RTIcap.AttributeNotOwned();
         eventCount++;
         ret = new EventRetractionHandle(theTime,eventCount,_fedId);
         rec.updateAttributeValues(theObject,theAttributes,theTime,theTag,fedAmb,_fedId);
       } catch(RTIcap.ObjectClassNotDefined ocnd) 
        { throw new RTIcap.ObjectNotKnown();}

       if(DEBUG) 
          System.out.println(" out of updateAttributeValuesWithTime() of ObjectManager ");

       return ret;

    }// end of updateAttributeValuesWithTime()

    /**
    * RTI 1.3
    */
    public void 
    updateAttributeValues(int theObject, 
           RTIcap.AttributeHandleValuePair[] theAttributes, 
           String theTag, RTIcap.FederateAmbassador fedAmb, short _fedId)
        throws RTIcap.ObjectNotKnown, RTIcap.AttributeNotDefined,
               RTIcap.AttributeNotOwned, RTIcap.FederateNotExecutionMember, 
               RTIcap.SaveInProgress, RTIcap.RestoreInProgress,  
               RTIcap.RTIinternalError
    {  if(DEBUG) 
          System.out.println(" in updateAttributeValues() of ObjectManager for "+theObject);
       FederateInfo fdInfo = fedEx.getFederateInfo(_fedId);
       if( fdInfo == null ) throw new RTIcap.FederateNotExecutionMember();
       try
       { updateAttributeValuesWithTime(theObject, theAttributes, 
                 fdInfo.getFederateTime(), theTag, fedAmb, _fedId);
       } catch( RTIcap.InvalidFederationTime e) {}
       if(DEBUG) 
          System.out.println(" out of updateAttributeValues() of ObjectManager ");
    }// end of updateAttributeValues()

    /**
    * Receive interaction from the federate
    * Find the interaction class
    * sendInteraction
    */
    public RTIcap.EventRetractionHandle 
    sendInteractionWithTime(short theInteraction, 
        RTIcap.ParameterHandleValuePair[] theParameters, double theTime, 
        String theTag, RTIcap.FederateAmbassador fedAmb, short theHandle)
        throws RTIcap.InteractionClassNotDefined, 
               RTIcap.InteractionClassNotPublished, 
               RTIcap.InteractionParameterNotDefined, 
               RTIcap.InvalidFederationTime, RTIcap.FederateNotExecutionMember, 
               RTIcap.SaveInProgress, RTIcap.RestoreInProgress, 
               RTIcap.RTIinternalError
    {  FederateInfo fdInf = fedEx.getFederateInfo(theHandle);
       if(fdInf==null) throw new RTIcap.FederateNotExecutionMember();

       RTIcap.EventRetractionHandle ret = null;

       InteractionRecord intRec = interActionNaming.getInteractionRecord(theInteraction);
       if(intRec==null) throw new RTIcap.InteractionClassNotDefined();
       if(!intRec.canPublish(theHandle)) throw new RTIcap.InteractionClassNotPublished();

       eventCount++;
       ret = new EventRetractionHandle(theTime,eventCount,theHandle);
       intRec.publishParams(theParameters,theTime,theTag,fdInf);
       return ret;
    }// end of sendInteractionWithTime()

    /**
    * Receive interaction from the federate
    * Find the interaction class
    * sendInteraction
    */
    public void 
    sendInteraction(short theInteraction, 
         RTIcap.ParameterHandleValuePair[] theParameters, String theTag, 
         RTIcap.FederateAmbassador fedAmb, short theHandle)
        throws RTIcap.InteractionClassNotDefined,
               RTIcap.InteractionClassNotPublished, 
               RTIcap.InteractionParameterNotDefined, 
               RTIcap.FederateNotExecutionMember, RTIcap.SaveInProgress, 
               RTIcap.RestoreInProgress, RTIcap.RTIinternalError
    {  /** Get federate Info */
       FederateInfo fdInf = fedEx.getFederateInfo(theHandle);
       if(fdInf==null) throw new RTIcap.FederateNotExecutionMember();
       /** Get Interaction Record */
       InteractionRecord intRec = interActionNaming.getInteractionRecord(theInteraction);
       if(intRec==null) throw new RTIcap.InteractionClassNotDefined();
       if(!intRec.canPublish(theHandle)) throw new RTIcap.InteractionClassNotPublished();
       /** Publish given parameters */
       intRec.publishParams(theParameters,fdInf.getFederateTime(),theTag,fdInf);
    }// end of sendInteraction()

    /**
    * 6.8
    */
    public RTIcap.EventRetractionHandle 
    deleteObject( int objectID, double theTime, String theTag, 
                  FederateAmbassador fedAmb, short _fedHandle)
         throws RTIcap.ObjectNotKnown, RTIcap.DeletePrivilegeNotHeld, 
                RTIcap.InvalidFederationTime, RTIcap.FederateNotExecutionMember,  
                RTIcap.ConcurrentAccessAttempted, RTIcap.RTIinternalError
    {

       isExecutionMember(_fedHandle); 
       if(DEBUG) System.out.println(" in deleteObject() for Object "+objectID);

       RTIcap.EventRetractionHandle ret = null;
       eventCount++;
       ret = new EventRetractionHandle(theTime,eventCount,_fedHandle);
       Short theClass = (Short) objIdToClassId.get(new Integer(objectID));

       try
       {
       if(theClass!=null)
       {
          /** Find the Record of this object */
          ObjectRecord rec = objNaming.getObjectRecord(theClass.shortValue()); 

          /** Register this object to this Class */
          rec.deregister_object(objectID,fedEx.getFederateInfo(_fedHandle));

          /** Keeps the mapping between ObjectID and ClassID */ 
          objIdToClassId.remove(new Integer(objectID));

       }// end of if
       } catch(RTIcap.ObjectClassNotDefined ocnd) 
         { throw new RTIcap.ObjectNotKnown(); }

       if(DEBUG) System.out.println(" out of deleteObject() ");

       return ret;

    }// end of deleteObject()

    /**
    * 6.11
    */
    public void 
    changeAttributeTransportType(int theObject, short[] theAttributes, 
         RTIcap.TransportType theType, RTIcap.FederateAmbassador fedAmb, 
         short _fedHandle)
         throws RTIcap.ObjectNotKnown, RTIcap.AttributeNotDefined, 
                RTIcap.AttributeNotOwned, RTIcap.InvalidTransportationHandle, 
                RTIcap.FederateNotExecutionMember, 
                RTIcap.SaveInProgress, RTIcap.RestoreInProgress,
                RTIcap.RTIinternalError
    {
       isExecutionMember(_fedHandle); 
    }// end of changeAttributeTransportType()

    /**
    * 6.12
    */
    public void 
    changeInteractionTransportType(short theClass, 
         RTIcap.TransportType theType, RTIcap.FederateAmbassador fedAmb, 
         short _fedHandle)
         throws RTIcap.InteractionClassNotDefined, 
                RTIcap.InteractionClassNotPublished, 
                RTIcap.InvalidTransportationHandle, 
                RTIcap.FederateNotExecutionMember, 
                RTIcap.SaveInProgress, RTIcap.RestoreInProgress,
                RTIcap.RTIinternalError
    {
       isExecutionMember(_fedHandle); 
    }// end of changeInteractionTransportType()

    /**
    * 6.15
    *  Update given attributes of the object instance
    */
    public void 
    requestObjectAttributeValueUpdate1(int theObject, short[] theAttributes , 
        RTIcap.FederateAmbassador fedAmb, short _fedId)
        throws RTIcap.ObjectNotKnown, RTIcap.AttributeNotDefined,
               RTIcap.FederateNotExecutionMember, RTIcap.SaveInProgress, 
               RTIcap.RestoreInProgress, RTIcap.RTIinternalError
    {  if(DEBUG) 
          System.out.println(" in requestObjectAttributeValueUpdate1() for object "+theObject);
       isExecutionMember(_fedId); 
       ObjectInstanceRecord oRec = objNaming.getObjectInstance(theObject); 
       ObjectRecord         rec  = oRec.getObjectClass();
       for(int i=0;i<theAttributes.length;i++) 
             rec.haveAttribute(theAttributes[i]);
       oRec.requestObjectAttributeValueUpdate(theAttributes);
       if(DEBUG) 
          System.out.println(" out of requestObjectAttributeValueUpdate1() ");
    }// end of requestObjectAttributeValueUpdate1()

    /**
    * 6.15
    *  Update all the attributes of the object instance
    */
    public void 
    requestObjectAttributeValueUpdate2(int theObject,
        RTIcap.FederateAmbassador fedAmb, short _fedId)
        throws RTIcap.ObjectNotKnown, RTIcap.FederateNotExecutionMember,
               RTIcap.SaveInProgress, RTIcap.RestoreInProgress, 
               RTIcap.RTIinternalError
    {
       if(DEBUG) 
          System.out.println(" in requestObjectAttributeValueUpdate2() for object "+theObject);
       isExecutionMember(_fedId); 
       ObjectInstanceRecord rec = objNaming.getObjectInstance(theObject); 
       rec.requestObjectAttributeValueUpdate(rec.getAllAttributeHandles());
       if(DEBUG) 
          System.out.println(" out of requestObjectAttributeValueUpdate2() ");
    }// end of requestObjectAttributeValueUpdate2()

    /**
    * This will cause a Provide Attribute Value Update method calls.
    *  Update given attributes of all object instances in this class
    */
    public void 
    requestClassAttributeValueUpdate1(short theClass, short[] theAttributes,
                                FederateAmbassador _fedHandle, short _fedId )
         throws RTIcap.ObjectClassNotDefined, RTIcap.AttributeNotDefined, 
                RTIcap.FederateNotExecutionMember, RTIcap.RTIinternalError
    {
       if(DEBUG) 
          System.out.println(" in requestClassAttributeValueUpdate1() for class "+theClass);
       isExecutionMember(_fedId); 
       /** Find the Record of this object */
       ObjectRecord rec = objNaming.getObjectRecord(theClass); 
       for(int i=0;i<theAttributes.length;i++) 
             rec.haveAttribute(theAttributes[i]);
       rec.requestClassAttributeValueUpdate(theAttributes);
       if(DEBUG) 
          System.out.println(" out of requestClassAttributeValueUpdate1() ");
    }// end of requestClassAttributeValueUpdate1()

    /**
    *  Update all the attributes of all object instances in this class
    */
    public void 
    requestClassAttributeValueUpdate2(short theClass,
        RTIcap.FederateAmbassador fedAmb, short _fedId)
        throws RTIcap.ObjectClassNotDefined, RTIcap.FederateNotExecutionMember,
               RTIcap.SaveInProgress, RTIcap.RestoreInProgress, RTIcap.RTIinternalError
    {  if(DEBUG) System.out.println(" in requestClassAttributeValueUpdate2() for class "+theClass);
       isExecutionMember(_fedId); 
       /** Find the Record of this object */
       ObjectRecord rec = objNaming.getObjectRecord(theClass); 
       try
       { rec.requestClassAttributeValueUpdate(rec.getAllAttributeHandles());
       } catch(RTIcap.AttributeNotDefined ad) {}
       if(DEBUG) System.out.println(" out of requestClassAttributeValueUpdate2() ");
    }// end of requestClassAttributeValueUpdate2()

    /**
    */
    public RTIcap.EventRetractionHandle 
    deleteObjectInstanceWithTime(int objectID, double theTime, String theTag, 
           RTIcap.FederateAmbassador fedAmb, short _fedHandle)
        throws RTIcap.ObjectNotKnown, RTIcap.DeletePrivilegeNotHeld,
               RTIcap.InvalidFederationTime, RTIcap.FederateNotExecutionMember,
               RTIcap.SaveInProgress, RTIcap.RestoreInProgress, RTIcap.RTIinternalError
    {  if(DEBUG) 
          System.out.println(" in deleteObjectInstanceWithTime() for Object "+objectID);
       isExecutionMember(_fedHandle); 
       deleteObjectInstance(objectID, theTag, fedAmb, _fedHandle);
       RTIcap.EventRetractionHandle ret = null;
       eventCount++;
       ret = new EventRetractionHandle(theTime,eventCount,_fedHandle);
       if(DEBUG) System.out.println(" out of deleteObjectInstanceWithTime() ");
       return ret;
    }// end of deleteObjectInstanceWithTime()

    /**
    */
    public void 
    deleteObjectInstance(int objectID, String theTag, 
        RTIcap.FederateAmbassador fedAmb, short _fedHandle)
        throws RTIcap.ObjectNotKnown, RTIcap.DeletePrivilegeNotHeld,
               RTIcap.FederateNotExecutionMember, RTIcap.SaveInProgress,
               RTIcap.RestoreInProgress, RTIcap.RTIinternalError
    {  if(DEBUG) 
          System.out.println(" in deleteObjectInstance() for Object "+objectID);
       isExecutionMember(_fedHandle); 

       ObjectInstanceRecord oRec = objNaming.getObjectInstance(objectID);
       if(oRec.getOwnerHandle()!=_fedHandle) throw new RTIcap.DeletePrivilegeNotHeld();
       ObjectRecord rec = oRec.getObjectClass();
       if(rec==null) throw new RTIcap.RTIinternalError();

       /** Remove this object from this Class */
       rec.deregister_object(objectID,fedEx.getFederateInfo(_fedHandle));

       /** Remote the mapping between ObjectID and ClassID */
       objIdToClassId.remove(new Integer(objectID));

       if(DEBUG) System.out.println(" out of deleteObjectInstance() ");

    }// end of deleteObjectInstance()

    /**
    */
    public void 
    localdeleteObjectInstance(int objectID, RTIcap.FederateAmbassador fedAmb, 
        short theHandle)
        throws RTIcap.ObjectNotKnown, RTIcap.FederateOwnsAttributes,
               RTIcap.FederateNotExecutionMember, RTIcap.SaveInProgress, 
               RTIcap.RestoreInProgress, RTIcap.RTIinternalError
    {
    }// end of localdeleteObjectInstance()


   /** */
   final void 
   isExecutionMember( short fedId ) throws RTIcap.FederateNotExecutionMember {
      if(fedEx.getFederateInfo(fedId)==null)
         throw new RTIcap.FederateNotExecutionMember();
   }// end of isExecutionMember

    /**
    */
    public
    void retract(RTIcap.EventRetractionHandle theHandle)
         throws RTIcap.InvalidRetractionHandle, 
                RTIcap.FederateNotExecutionMember, 
                RTIcap.ConcurrentAccessAttempted, RTIcap.RTIinternalError, 
                RTIcap.Deadlock
    {
    }

}// end of ObjectManager

