/**
*
*/
package owrti;
import owrti.evt.*;
import RTIcap.*;
import java.util.*;
import java.io.*;

/**
*
*  ObjectRecord object keeps information about one particular object
*  class. It knows its place in the Object Class Hierarchy. It knows 
*  its instances and its attributes.
*
* @date April 17, 1998 ( Last Updated March 31, 1999)
* @author H. Timucin Ozdemir
*
*/
public class 
ObjectRecord  extends RegionSubscriberRegistry implements RTIDebug
{
   /** Id of the object. It came as an short value. Might be an object */
   short objectId;

   /** Name of the object. It contains all the predecesors */
   String objectClassName;

   /** If root, it is null. */
   ObjectRecord father;                // Father in the Object Tree

   /** Attribute Information center */
   AttributeNaming attributeDirectory;   

   /** */
   Hashtable attWatcher = new Hashtable();

   /** Object Instances are going to be handled here */
   Hashtable objFactory = new Hashtable();

   //RegionInfo defRegion;
   //Hashtable  regions = new Hashtabel();

   /** In which federation this object exists */
   FederationExecution fedEx; 

   RTIMsgDispatcher dispatcher = new RTIMsgDispatcher();

   /** */
   protected ObjectRecord() {  
      attributeDirectory = new AttributeNaming(this); 
   }// end of ObjectRecord

   /** */
   public ObjectRecord(FederationExecution _fedEx) {  
      attributeDirectory = new AttributeNaming(this); 
      fedEx = _fedEx;
   }// end of ObjectRecord

   /** */
   public ObjectRecord(FederationExecution _fedEx, ObjectRecord _father) { 
       father = _father;
       attributeDirectory = new AttributeNaming(this); 
       fedEx = _fedEx;
   }// end of ObjectRecord

   /**
   * Returns the name of the attribute.
   */
   public String 
   getAttributeName(short theHandle) throws RTIcap.AttributeNotDefined {
       return attributeDirectory.getAttributeName(theHandle);
   }// end of getAttributeName()

   /**
   * returns the id of the attribute.
   */
   public short getAttributeHandle(String theName) throws RTIcap.NameNotFound {
       System.out.println(" Search for " +theName+" in getAttributeHandle() " );
       return attributeDirectory.getAttributeHandle(theName);
   }// end of getAttributeHandle()

   /**
   * returns the id of the attribute.
   */
   public short[] getAllAttributeHandles() {  
      return attributeDirectory.getAllHandles();
   }// end of getAllAttributeHandles()

   /** */
   public boolean haveAttribute(short theHandle) throws RTIcap.AttributeNotDefined { return attributeDirectory.haveAttribute(theHandle);
   }// end of haveAttribute()


   /** */
   public String getName() { return objectClassName; }

   /** */
   public short getId() { return objectId; }

   /** */
   public void setName(String name) { objectClassName=name; }

   /** */
   public void setId(short _id) { objectId=_id; }

   /** 
   * Actually we need more info about the Attribute.
   * But, right now, this will do it.
   */
   public void
   addAttribute(String name) 
   { 
      if(DEBUG) System.out.println(" name : "+name); 

      if(name!=null) 
      {   
         short ret = attributeDirectory.addAttribute(name);

         /*******************************************************/
         /** For Class attribute definitions, there is no owner */
         /*******************************************************/
         if(ret>0) 
           attWatcher.put(new Short(ret),new AttributeWatcher(this,ret,name,(short)-1));

      }// end of if

   }// end of addAttribute()

   /** */
   public AttributeNaming 
   getAttributeNaming(short theHandle) throws RTIcap.RTIinternalError {
       return attributeDirectory;
   }// end of getAttributeNaming()

   /** */
   public AttributeWatcher 
   getAttributeWatcher(short theAttHandle) throws RTIcap.RTIinternalError {
       return (AttributeWatcher)attWatcher.get(new Short(theAttHandle));
   }// end of getAttributeWatcher()

   /**
   *  Deletes an object and broadcast this event to everybody.
   *
   *  1. Find the ObjectInstanceRecord
   *  2. Tell everybody to delete this object from their locals
   *  3. Remove this object from the object factory
   *  4. Remove this object from all Attribute Watchers. 
   *
   */
   public void
   deregister_object( int theObject, FederateInfo _fedHandle )
   {
      ObjectInstanceRecord rec = (ObjectInstanceRecord) objFactory.get(new Integer(theObject));
      if(rec!=null)
      {
         /******************************************************/
         /** Inform everybody who has the proxy of this object */
         /******************************************************/
         if(_fedHandle!=null) 
            dispatcher.addRTIMsg(new BroadcastRemoveObject(theObject,_fedHandle,fedEx));
         objFactory.remove(new Integer(theObject));

         /********************************/
         /** Remove it from all Watchers */
         /********************************/
         Enumeration e = attWatcher.elements();
         AttributeWatcher curr;

         for(;e.hasMoreElements();)
         {
             curr = (AttributeWatcher) e.nextElement();

             if(curr!=null) curr.deregister_subscriber(_fedHandle.getFederateID());
           
         }// end of for
         

      }// end of if

   }// end of deregister_object()

   /**
   *  Defines an object and broadcast this event to everybody.
   */
   public void register_object( int theObject, String theName, 
                                FederateInfo   _fedHandle, 
                                short _fedId,  FederationExecution fedEx )
   {
      register_object(theObject,theName,_fedHandle,_fedId,fedEx,getDefaultRegion());
   }// end of register_object()

   /**
   *  Defines an object and broadcast this event to everybody.
   */
   public void register_object( int theObject, String theName, 
                                FederateInfo   _fedHandle, 
                                short _fedId,  FederationExecution fedEx,
                                RegionInfo regRegion )
   {
      ObjectInstanceRecord rec = null;
      if(theName==null) 
           rec = new ObjectInstanceRecord(_fedId,objectId,theObject,_fedHandle,this,dispatcher,regRegion );
      else rec = new ObjectInstanceRecord(_fedId,objectId,theObject,_fedHandle,this,dispatcher,regRegion,theName);

      objFactory.put(new Integer(theObject), rec);

      if(_fedHandle!=null) 
         dispatcher.addRTIMsg(new BroadcastDiscoverObject(theObject,objectId,getAllAttributeHandles(),_fedHandle,fedEx));

   }// end of register_object()

   /**
   *  Inform the Federate about existance of other object instances.
   */
   public void broadcastObjectInstances( FederateInfo _fedHandle ) {
      dispatcher.addRTIMsg(new BroadcastObjectInstances(objFactory.elements(),_fedHandle));
   }// end of broadcastObjectInstances()

   /**
   *  Broadcast everybody the update message about this particular object
   *  and its attributes. Should I keep copy of these attributes?
   */
   public synchronized void
   updateAttributeValues( int theObject, 
                          RTIcap.AttributeHandleValuePair[] theAttributes,
                          double theTime, String theTag,
                          FederateAmbassador _fedHandle, short _fedId )

         throws RTIcap.AttributeNotDefined,
                RTIcap.AttributeNotOwned, RTIcap.InvalidFederationTime,
                RTIcap.FederateNotExecutionMember, RTIcap.RTIinternalError
   { dispatcher.addRTIMsg(new BroadcastUpdateAttributeValues(theObject,theAttributes,theTime,theTag,_fedHandle,_fedId));
   }// end of updateAttributeValues()

   /**
   * Ask each federate who has an instance of this object class to update
   * their attributes.
   */
   public void requestClassAttributeValueUpdate(short[] theAttributes )
         throws RTIcap.AttributeNotDefined, RTIcap.FederateNotExecutionMember,
                RTIcap.RTIinternalError
   { dispatcher.addRTIMsg(new BroadcastRequestClassValueUpdate(objFactory.elements(),theAttributes));
   }// end of requestClassAttributeValueUpdate()

   /** */
   public ObjectInstanceRecord getInstance( int theHandle ) 
   { return (ObjectInstanceRecord) objFactory.get(new Integer(theHandle)); } 

   /*********************************/
   /** Data Distribution Management */
   /*********************************/

   /**
   *  send the same message to all object instances of this class.
   */
   public void
   requestUpdateAttributesInRegion( short[] attributes, RegionInfo regInf) { 
     dispatcher.addRTIMsg(new BroadcastRequestClassValueUpdate(objFactory.elements(),attributes,regInf));
   }// end of requestUpdateAttributesInRegion()

   /**
   * 1. Find AttWatcher
   * 2. Add this region to AttributeWatcher Subscribing regions list.
   */ 
   public void
   subscribeAttributesWithRegion( short[] attributeList, RegionInfo rgn)
   {
      AttributeWatcher att;

      for(int i=0;i<attributeList.length;i++)
      {
         try 
         {  
            att = getAttributeWatcher(attributeList[i]);

            if( att != null ) att.register_subscribingRegion(rgn); 

         } catch( RTIcap.RTIinternalError ie) {}

      }// end of for

   }// end of subscribeAttributesWithRegion()

   /**
   * Returns default region definition for the attribute.
   */
   public RegionInfo
   getDefaultRegionForAttribute( short attribute )
   {
      RegionInfo ret = null;
      AttributeWatcher att;

      try 
      {  
          att = getAttributeWatcher(attribute);
         if( att != null ) ret = att.getRegionInfo();

      } catch(RTIcap.RTIinternalError ie) {}

      return ret;

   }// end of getDefaultRegionForAttribute()

   /**
   * Returns the handle of all owners of given attribute
   */
   public Enumeration
   getAllOwnersOfAttribute( short theHandle )
   {
      Vector fdIds = new Vector();
      Enumeration e = objFactory.elements();
      ObjectInstanceRecord rec;
      int ownId;
      Short ky;

      Vector ret = new Vector();
      FederateInfo n;

      for(;e.hasMoreElements();)
      {
         rec = (ObjectInstanceRecord) e.nextElement();

         if( rec != null )
         {
            ownId = rec.getAttributeOwner(theHandle);
            if(ownId >0 && !fdIds.contains(new Integer(ownId)))
               fdIds.addElement(new Integer(ownId));
               
         }// end of if

      }// end of for

      for(int i=0;i<fdIds.size();i++)
      {
         ky = (Short) ret.elementAt(i);
         if( ky != null )
         {
            n = fedEx.getFederateInfo(ky.shortValue());
            if(n!=null) ret.addElement(n); 

         }// end of if

      }// end of foR

      return ret.elements();

   }// end of getAllOwnersOfAttribute()

   /**
   *  Sends "turnUpdatesOnForObjectinstance" message to all owners of the
   *  particular attribute of an object instance.
   */
   void turnUpdatesOnOffForAllOwners(short theHandle, boolean OnOff) {
      dispatcher.addRTIMsg(new BroadcastTurnUpdatesOnOffForAllOwners(objFactory.elements(),fedEx,theHandle,OnOff));
   }// end of turnUpdatesOnOffForAllOwners()

   /** */
   public void dump(OutputStream out) {
      System.out.println(" Object Name : "+objectClassName);
      System.out.println(" Object Id   : "+objectId);
      int numOfAtt = attributeDirectory.getNumberOfAttributes(); 
      System.out.println(" # of Attributes : "+numOfAtt);
      for(int i=1;i<(numOfAtt+1);i++)
      try {
         System.out.println(i+"th att name : "+attributeDirectory.getAttributeName((short)i));
      } catch(RTIcap.AttributeNotDefined  and) {}
   }// end of dump()

   public void dump2(OutputStream out) {
      try 
      { PrintWriter pw = new PrintWriter(out,true);
        pw.println(" Object Name : ");
        pw.println(" Object Id   : ");
        int numOfAtt = attributeDirectory.getNumberOfAttributes(); 
        pw.println(" # of Attributes : "+numOfAtt);
        for(short i=0;i<numOfAtt;i++)
          pw.println(i+"th att name : "+attributeDirectory.getAttributeName(i));
      } catch(RTIcap.AttributeNotDefined  and) {}
   }// end of dump()

   /**
   *  Broadcasts UpdateAttributeValues event to all federates in the 
   *  Federation.
   */
   class BroadcastUpdateAttributeValues implements RTIMessage, RTIDebug
   {
     int theObject;
     RTIcap.AttributeHandleValuePair[] theAttributes;
     double theTime;
     String theTag;
     FederateAmbassador _fedAmb;
     short  _fedId;
   
     BroadcastUpdateAttributeValues( int theObject, 
                                     RTIcap.AttributeHandleValuePair[] theAttributes,
                                     double theTime, String theTag,
                                     FederateAmbassador _fedAmb, short  _fedId)
     {
        this.theObject=theObject;
        this.theAttributes=theAttributes;
        this.theTime=theTime;
        this.theTag=theTag;
        this._fedAmb=_fedAmb;
        this._fedId=_fedId;
     }// end of CONSTRUCTOR

     public void execute() { 
       if(DEBUG) System.out.println(" In updateAttributeValues() for "+theObject ) ;
       if(DEBUG) System.out.println(" theAttributes.length "+theAttributes.length ) ;
       for(int i=0;i<theAttributes.length;i++) {
         AttributeWatcher attWatch = (AttributeWatcher)attWatcher.get(new Short(theAttributes[i].handle));
         if(attWatch!=null) 
          try
          {
            attWatch.broadcast(theObject,theAttributes[i],theTime,theTag,_fedAmb,_fedId);
          } catch(RTIcap.FederateInternalError fie) {}
            catch(RTIcap.ObjectNotKnown onk) {}
            catch(RTIcap.AttributeNotKnown ank) {}
            catch(RTIcap.InvalidFederationTime ank2) {}

       }// end of for
       if(DEBUG) 
          System.out.println(" Out of updateAttributeValues() for "+theObject ) ;
     }// end of execute()
    
   }// end of class BroadcastUpdateAttributeValues

}// end of class ObjectRecord

/**
*  Broadcasts RemoveObjectInstance event to all federates in the 
*  Federation.
*/
class BroadcastRemoveObject implements RTIMessage, RTIDebug {
   int theObject;
   FederateInfo _fedHandle;
   FederationExecution fedEx;
   
   BroadcastRemoveObject( int theObject, FederateInfo _fedHandle, FederationExecution fedEx)
   {
      this.theObject=theObject;
      this._fedHandle=_fedHandle;
      this.fedEx=fedEx;
   }// end of CONSTRUCTOR

   public void execute() {
      if(DEBUG) System.out.println(" In broadcastRemoveObject() " );
      Enumeration e = fedEx.getAllFederateInfos();
      if(e!=null&&e.hasMoreElements()) {
         if(DEBUG) System.out.println(" e is not NULL " );
         FederateInfo n;
         double time = 10;
         String tag  = new String("remove");
         RTIcap.EventRetractionHandle retract = new RTIcap.EventRetractionHandle(time,0,_fedHandle.getFederateID());
         if(DEBUG) System.out.println(" before for " );
         for(;e.hasMoreElements();) { 
            n = (FederateInfo) e.nextElement();
            if(DEBUG) System.out.println(" FederateID : "+n.getFederateID() );
            if( n!=null && _fedHandle.getFederateID()!=n.getFederateID() ) 
               n.addRTIEvent(new EvtRemoveObjectInstance(theObject,tag,retract));
         }// end of for
      }// end of if
      if(DEBUG) System.out.println(" Out Of broadcastRemoveObject() " );
   }// end of execute()
    
}// end of class BroadcastRemoveObject

/**
*  Inform all federates in the federation execution about the 
*  new object definition.
*/
class BroadcastDiscoverObject implements RTIMessage, RTIDebug
{
   short[] attributes;
   int theObject;
   short classId;
   FederateInfo _fedHandle;
   FederationExecution fedEx;
   
   BroadcastDiscoverObject( int theObject, short classId, short[] attributes, 
                            FederateInfo _fedHandle, FederationExecution fedEx)
   {
      this.theObject=theObject;
      this.attributes=attributes;
      this.classId=classId;
      this._fedHandle=_fedHandle;
      this.fedEx=fedEx;
   }// end of CONSTRUCTOR

   public void execute() {
      if(DEBUG) System.out.println(" In broadcastDiscoverObject() " );
      Enumeration e = null;
      if(fedEx!=null) e = fedEx.getAllFederateInfos();
      if(e!=null&&_fedHandle!=null&&e.hasMoreElements()) {
         if(DEBUG) System.out.println(" e is not NULL " );
         FederateInfo n;
         double time = 10;
         String tag  = new String("discover");
         RTIcap.EventRetractionHandle retract = new RTIcap.EventRetractionHandle(time,0,_fedHandle.getFederateID());
         if(DEBUG) System.out.println(" before for " );
         for(;e.hasMoreElements();) { 
            n = (FederateInfo) e.nextElement();
            if(DEBUG) System.out.println(" FederateID : "+n.getFederateID() );
            if( n!=null && _fedHandle.getFederateID()!=n.getFederateID() ) 
                n.addRTIEvent(new EvtDiscoverObject(theObject,classId,_fedHandle.getFederateID(),time));
         }// end of for
         _fedHandle.addRTIEvent( new EvtStartRegistrationForObjectClass(classId));
         if(fedEx.getNumberOfFederates()>1)
           _fedHandle.addRTIEvent( new EvtTurnUpdatesOnForObjectInstance(theObject,attributes));
         else   
           _fedHandle.addRTIEvent( new EvtTurnUpdatesOffForObjectInstance(theObject,attributes));
      }// end of if
      if(DEBUG) System.out.println(" Out Of broadcastDiscoverObject() " );
   }// end of execute()
    
}// end of class BroadcastDiscoverObject

/**
*  Inform the Federate about existance of other object instances.
*/
class BroadcastObjectInstances implements RTIMessage, RTIDebug
{
   Enumeration e;
   FederateInfo _fedHandle;
   
   BroadcastObjectInstances( Enumeration e, FederateInfo _fedHandle ) { 
      this.e=e;
      this._fedHandle=_fedHandle;
   }// end of CONSTRUCTOR

   public void execute() {
      if(DEBUG) System.out.println(" In broadcastObjectInstances() " );

      if( e != null) {
         for(;e.hasMoreElements();) { 
            ObjectInstanceRecord rec = (ObjectInstanceRecord) e.nextElement();
            if( rec!=null && rec.getOwnerHandle()!= _fedHandle.getFederateID() )
             _fedHandle.addRTIEvent(new EvtDiscoverObject(rec.getObjectInstanceHandle(), rec.getObjectClassHandle(), rec.getOwnerHandle() ));
         }// end of for
      }// end of if

      if(DEBUG) System.out.println(" Out Of broadcastObjectInstances() " );

   }// end of execute()
    
}// end of class BroadcastObjectInstances

/**
* Ask each federate who has an instance of this object class to update
* their attributes.
*/
class BroadcastRequestClassValueUpdate implements RTIMessage, RTIDebug
{
   Enumeration e;            //ObjectInstanceRecords
   short[] theAttributes;
   RegionInfo regInf=null;

   BroadcastRequestClassValueUpdate( Enumeration e, short[] theAttributes ) { 
      this(e,theAttributes,null);
   }// end of CONSTRUCTOR

   BroadcastRequestClassValueUpdate( Enumeration e, short[] theAttributes, RegionInfo regInf ) { 
      this.e=e;
      this.theAttributes=theAttributes;
      this.regInf=regInf;
   }// end of CONSTRUCTOR

   public void execute() {
      if( theAttributes!=null && theAttributes.length>0 && e!=null &&
          e.hasMoreElements() ) {
        ObjectInstanceRecord objInstance;
        for(;e.hasMoreElements();) { 
            /** Get the Object instance Record */ 
            objInstance = (ObjectInstanceRecord) e.nextElement();
            if(objInstance!=null)  
               objInstance.requestUpdateAttributesInRegion2(theAttributes,regInf); 
        }// end of for
      }// end of if
      
   }// end of execute()
    
}// end of class BroadcastRequestClassValueUpdate


/**
*  Sends "turnUpdatesOnForObjectinstance" message to all owners of the
*  particular attribute of an object instance.
*/
class BroadcastTurnUpdatesOnOffForAllOwners implements RTIMessage, RTIDebug
{
   Enumeration e;            //ObjectInstanceRecords
   short theHandle;
   FederationExecution fedEx;
   boolean OnOff;
   RegionInfo rgnInf;

   BroadcastTurnUpdatesOnOffForAllOwners( Enumeration e, FederationExecution fedEx, short theHandle, boolean OnOff ) { 
      this(e,fedEx,theHandle,OnOff,null);
   }// end of CONSTRUCTOR

   BroadcastTurnUpdatesOnOffForAllOwners( Enumeration e, FederationExecution fedEx, short theHandle, boolean OnOff, RegionInfo rgnInf ) { 
      this.e=e;
      this.theHandle=theHandle;
      this.fedEx=fedEx;
      this.rgnInf=rgnInf;
   }// end of CONSTRUCTOR

   public void execute() {
      ObjectInstanceRecord rec;
      short[] attribs = new short[1]; attribs[0]=theHandle;
      for(;e.hasMoreElements();) {
         rec = (ObjectInstanceRecord) e.nextElement();
         if( rec != null ) rec.turnUpdatesOnOff(attribs,OnOff,rgnInf);
      }// end of for
   }// end of execute()
    
}// end of class BroadcastTurnUpdatesOnOffForAllOwners

