/**
*  RTI 1.3
*/
package owrti;
import owrti.evt.*;
import RTIcap.*;
import npacrti.*;
import java.util.*;

/*
*
* @date April 20, 1998 ( Updated at May 06, 1998 )
* @author H. Timucin Ozdemir 
*
*/
public class 
OwnershipManager extends _OwnershipManagerInterfaceImplBase implements RTIDebug
{
      FederationExecution  fedEx;
        ObjectClassNaming  objNaming; 
    RequestOwnershipTable  ownersTable;

    /**
    */
    public
    OwnershipManager(FederationExecution _fedEx, ObjectClassNaming _objNaming )
    {  
       fedEx       = _fedEx; 
       objNaming   = _objNaming;
       ownersTable = new RequestOwnershipTable(fedEx);

    }// end of CONSTRUCTOR

    /**
    * 7.2
    */
    public void 
    unconditionalAttributeOwnershipDivestiture( int object, short[] attributes,
                                                short theHandle)
        throws RTIcap.AttributeNotDefined, RTIcap.AttributeNotOwned,
               RTIcap.FederateNotExecutionMember, RTIcap.ObjectNotKnown,
               RTIcap.RestoreInProgress, RTIcap.RTIinternalError, 
               RTIcap.SaveInProgress
    {  isExecutionMember(theHandle);
        ObjectInstanceRecord instance = objNaming.getObjectInstance(object);
        if(instance==null) throw new RTIcap.ObjectNotKnown();

        instance.releaseOwnerships(attributes,true);

    }// end of unconditionalAttributeOwnershipDivestiture()

    /**
    * 7.3
    */
    public void 
    negotiatedAttributeOwnershipDivestiture( int object, short[] attributes,
                                             String tag, short theHandle )
        throws RTIcap.AttributeAlreadyBeingDivested, RTIcap.AttributeNotDefined,
               RTIcap.AttributeNotOwned, RTIcap.FederateNotExecutionMember, 
               RTIcap.ObjectNotKnown, RTIcap.RestoreInProgress, 
               RTIcap.RTIinternalError, RTIcap.SaveInProgress
    {   isExecutionMember(theHandle);
        ObjectInstanceRecord instance = objNaming.getObjectInstance(object);
        if(instance==null) throw new RTIcap.ObjectNotKnown();

        instance.releaseOwnerships(attributes,false);

    }// end of negotiatedAttributeOwnershipDivestiture()

    /**
    * 7.7
    */
    public void 
    attributeOwnershipAcquisition( int object, short[] attributes, String tag,
                                   short theHandle)
        throws RTIcap.AttributeNotDefined, RTIcap.AttributeNotPublished,
               RTIcap.FederateNotExecutionMember, RTIcap.FederateOwnsAttributes,
               RTIcap.ObjectClassNotPublished, RTIcap.ObjectNotKnown, 
               RTIcap.RestoreInProgress, RTIcap.RTIinternalError, 
               RTIcap.SaveInProgress
    {  isExecutionMember(theHandle);
        FederateInfo fedInfo2;
        /** Get federate info object */
        FederateInfo fedInfo = (FederateInfo) fedEx.getFederateInfo(theHandle);
        if(fedInfo==null) throw new RTIcap.FederateNotExecutionMember();

        /** Get Object Instance Record */
        ObjectInstanceRecord instance = objNaming.getObjectInstance(object);
        if(instance==null) throw new RTIcap.ObjectNotKnown();

        AttributeInfo attInfo;

        /** For each attributes in attributes */
        for(int i=0;i<attributes.length;i++)
        {
           attInfo = instance.getAttributeInfo(attributes[i]);
 
           if(attInfo!=null) 
           {
              if(attInfo.getOwnerId() != -1 )
              {
                 fedInfo2 = (FederateInfo) fedEx.getFederateInfo(attInfo.getOwnerId());
                 if(fedInfo2!=null)
                 {
                    attInfo.setOwnershipAcquisitionInProgress();
                    fedInfo2.addRTIEvent(new EvtRequestAttributeOwnershipRelease(object,attributes[i],tag,theHandle,attInfo)); 
                 }// end of if

              }// end of if

           }// end of if
         
        }// end of for
  
    }// end of attributeOwnershipAcquisition()

    /**
    * 7.8
    * 1. Find the attribute
    * 2. If nobody owns it or transfer is possible, then
    *    2.a. inform requester with AttributeOwnershipAcquisitionNotification
    *    2.b. inform the owner with AttributeOwnershipDivestitureNotification
    * 3. Else: inform requester with AttributeOwnershipUnavailable
    */
    public void 
    attributeOwnerhsipAcquisitionIfAvailable( int object, short[] attributes,
                                              short theHandle)
        throws RTIcap.AttributeAlreadyBeingAcquired, RTIcap.AttributeNotDefined,
               RTIcap.AttributeNotPublished, RTIcap.FederateNotExecutionMember, 
               RTIcap.FederateOwnsAttributes, RTIcap.ObjectClassNotPublished, 
               RTIcap.ObjectNotKnown, RTIcap.RestoreInProgress, 
               RTIcap.RTIinternalError, RTIcap.SaveInProgress
    {  isExecutionMember(theHandle);
        /** Get federate info object */
        FederateInfo fedInfo = (FederateInfo) fedEx.getFederateInfo(theHandle);
        if(fedInfo==null) throw new RTIcap.FederateNotExecutionMember();

        /** Get Object Instance Record */
        ObjectInstanceRecord instance = objNaming.getObjectInstance(object);
        if(instance==null) throw new RTIcap.ObjectNotKnown();

        AttributeInfo attInfo;

        /** For each attributes in attributes */
        for(int i=0;i<attributes.length;i++)
        {
           attInfo = instance.getAttributeInfo(attributes[i]);
 
           if(attInfo!=null) {
              if(attInfo.getOwnerId() == -1 || attInfo.getOwnershipTransferPermission() )
              {
                 attInfo.setOwnershipAcquisitionInProgress();
                 short[] aa = new short[1];
                 aa[0]=attributes[i];
                 fedInfo.addRTIEvent(new EvtAttributeOwnershipAcquisitionNotification(object,aa)); 
                 attInfo.setOwnerId(theHandle);
                 attInfo.resetOwnershipTransferPermission();
                 attInfo.setOwned();
                 FederateInfo fedOwner = (FederateInfo) fedEx.getFederateInfo(attInfo.getOwnerId());
              
                 if(fedOwner!=null) fedOwner.addRTIEvent(new EvtAttributeOwnershipDivestitureNotification(object,attributes[i],attInfo.getOwnerId()));
                 attInfo.resetOwnershipAcquisitionInProgress();
              }
              else
                 fedInfo.addRTIEvent(new EvtAttributeOwnershipUnavailable(object,attributes[i],theHandle)); 

           }// end of if
         
        }// end of for

    }// end of attributeOwnerhsipAcquisitionIfAvailable()

    /**
    * 7.11
    */
    public short[] 
    attributeOwnershipReleaseResponse( int object, short[] attributes, 
                                       short theHandle )
        throws RTIcap.AttributeNotDefined, RTIcap.AttributeNotOwned,
               RTIcap.FederateNotExecutionMember,
               RTIcap.FederateWasNotAskedToReleaseAttribute, 
               RTIcap.ObjectNotKnown, RTIcap.RestoreInProgress,
               RTIcap.RTIinternalError, RTIcap.SaveInProgress
    {  isExecutionMember(theHandle);
       Vector processed = new Vector();

       /** Get federate info object */
       FederateInfo fedInfo = (FederateInfo) fedEx.getFederateInfo(theHandle);
       if(fedInfo==null) throw new RTIcap.FederateNotExecutionMember();

       /** Get Object Instance Record */
       ObjectInstanceRecord instance = objNaming.getObjectInstance(object);
       if(instance==null) throw new RTIcap.ObjectNotKnown();

       AttributeInfo attInfo;

       /** For each attributes in attributes */
       for(int i=0;i<attributes.length;i++) {
           attInfo = instance.getAttributeInfo(attributes[i]);
           if( attInfo != null ) {
              FederateInfo requester = ownersTable.getAttributeRequester(object,attributes[i]);
              if(requester!=null) {
                 ownersTable.deregister(requester.getFederateID(),object,attributes[i]);
                 requester.addRTIEvent(new EvtAttributeOwnershipDivestitureNotification(object,attributes[i],theHandle)); 
                 processed.addElement(new Short(attributes[i]));
                 attInfo.resetOwnershipTransferPermission() ;
                 attInfo.setOwnerId(requester.getFederateID()) ;
                 attInfo.resetOwnershipAcquisitionInProgress() ;
                 attInfo.setOwned() ;
              } else attInfo.setOwnershipTransferPermission() ;
           }// end of if
       }// end of for

       int j=0;
       short[] ret = null;
       if(processed.size()>0) { 
          ret = new short[processed.size()];
          Short val;
          for(int i=0;i<processed.size();i++) {
             val = (Short) processed.elementAt(i);
             if(val!=null) ret[j++] = val.shortValue(); 
          }// end of for
          if(j!=processed.size()) {
             short[] buf = new short[j];
             for(int i=0;i<j;i++) buf[j] = ret[j];
             ret = buf;
          }// end of if
       }// end of if
  
       return ret;

    }// end of attributeOwnershipReleaseResponse()

    /**
    * 7.12
    */
    public void 
    cancelNegotiatedAttributeOwnershipDivestiture( int object, 
                                 short[] attributes, short theHandle)
        throws RTIcap.AttributeDivestitureWasNotRequested,
               RTIcap.AttributeNotDefined, RTIcap.AttributeNotOwned, 
               RTIcap.FederateNotExecutionMember, RTIcap.ObjectNotKnown, 
               RTIcap.RestoreInProgress, RTIcap.RTIinternalError, 
               RTIcap.SaveInProgress
    {   isExecutionMember(theHandle);

        /** Get federate info object */
        FederateInfo fedInfo = (FederateInfo) fedEx.getFederateInfo(theHandle);
        if(fedInfo==null) throw new RTIcap.FederateNotExecutionMember();

        /** Get Object Instance Record */
        ObjectInstanceRecord instance = objNaming.getObjectInstance(object);
        if(instance==null) throw new RTIcap.ObjectNotKnown();

        AttributeInfo attInfo;

        /** For each attributes in attributes */
        for(int i=0;i<attributes.length;i++)
        {
           attInfo = instance.getAttributeInfo(attributes[i]);
 
           /************************************************************/
           /** If attribute exists and belongs to this federate, then  */
           /** cancel transfer permission issued before.               */
           /************************************************************/
           if( attInfo != null && attInfo.getOwnerId() == theHandle &&
               !attInfo.getOwnershipAcquisitionInProgress() ) 
               { attInfo.resetOwnershipTransferPermission() ; attInfo.setOwned(); }

        }// end of for
  
    }// end of cancelNegotiatedAttributeOwnershipDivestiture()

    /**
    * 7.13
    */
    public void 
    cancelAttributeOwnershipAcquisition(int object, short[] attributes, 
           short theHandle)
        throws RTIcap.AttributeAcquisitionWasNotRequested,
               RTIcap.AttributeAlreadyOwned, RTIcap.AttributeNotDefined, 
               RTIcap.FederateNotExecutionMember, RTIcap.ObjectNotKnown, 
               RTIcap.RestoreInProgress, RTIcap.RTIinternalError, 
               RTIcap.SaveInProgress
    {  int i;
       isExecutionMember(theHandle);
       Vector notInterested = new Vector();
       Vector Interested = new Vector();

       for(i=0;i<attributes.length;i++)
         if(!ownersTable.isFederateInterested(theHandle,object,attributes[i]))
            notInterested.addElement(new Short(attributes[i]));
         else Interested.addElement(new Short(attributes[i]));
       if(Interested.size()==attributes.length) 
          ownersTable.deregister(theHandle,object,attributes);
       else {
          short[] att = toShortArray(Interested);
          if(att!=null&&att.length>0) ownersTable.deregister(theHandle,object,att); 
       }
       if(notInterested.size()>0) {
          short[] att = toShortArray(notInterested);
          String mess = new String("Attributes ");
          for(i=0;i<att.length;i++) mess += (","+att[i]);
          mess += " are not requested!";
          throw new RTIcap.AttributeAcquisitionWasNotRequested(0,mess); 
       } 

    }// end of cancelAttributeOwnershipAcquisition()

    /**
    * 7.15
    */
    public void 
    queryAttributeOwnership(int theObject, short theAttribute, short theHandle)
         throws RTIcap.ObjectNotKnown, RTIcap.AttributeNotDefined, 
                RTIcap.AttributeNotDefined, RTIcap.FederateNotExecutionMember, 
                RTIcap.RTIinternalError
    {   isExecutionMember(theHandle);
       /** Get federate info object */
       FederateInfo fedInfo = (FederateInfo) fedEx.getFederateInfo(theHandle);
       if(fedInfo==null) throw new RTIcap.FederateNotExecutionMember();

       /** Get Object Instance Record */
       ObjectInstanceRecord instance = objNaming.getObjectInstance(theObject);
       if(instance==null) throw new RTIcap.ObjectNotKnown();

       /** Get Attribute Record */
       AttributeInfo att = instance.getAttributeInfo(theAttribute);
       if(att==null) throw new RTIcap.AttributeNotDefined();

       /** schedule informAttributeOwnership event */
       fedInfo.addRTIEvent(new EvtInformAttributeOwnership(theObject,theAttribute,att.getOwnerId()));  

    }// end of queryAttributeOwnership()

    /**
    * 7.17
    */
    public RTIcap._Boolean 
    isAttributeOwnedByFederate(int theObject, short theAttribute, short theHandle)
        throws RTIcap.AttributeNotDefined, RTIcap.FederateNotExecutionMember, 
               RTIcap.ObjectNotKnown, RTIcap.RestoreInProgress, 
               RTIcap.RTIinternalError, RTIcap.SaveInProgress
    {  isExecutionMember(theHandle);
       RTIcap._Boolean ret = RTIcap._Boolean.RTI_FALSE;

       /** Get federate info object */
       FederateInfo fedInfo = (FederateInfo) fedEx.getFederateInfo(theHandle);
       if(fedInfo==null) throw new RTIcap.FederateNotExecutionMember();

       /** Get Object Instance Record */
       ObjectInstanceRecord instance = objNaming.getObjectInstance(theObject);
       if(instance==null) throw new RTIcap.ObjectNotKnown();

       /** Get Attribute Record */
       AttributeInfo att = instance.getAttributeInfo(theAttribute);
       if(att==null) throw new RTIcap.AttributeNotDefined();

       if(theHandle==att.getOwnerId()) ret = RTIcap._Boolean.RTI_TRUE;

       return ret;

    }// end of isAttributeOwnedByFederate()

    final void
    isExecutionMember( short fedId ) throws RTIcap.FederateNotExecutionMember {
      if(fedEx.getFederateInfo(fedId)==null)
         throw new RTIcap.FederateNotExecutionMember();
    }// end of isExecutionMember

    final short[]
    toShortArray(Vector processed) {
      short[] ret = null;
      int j=0;
      if(processed.size()>0) {
          ret = new short[processed.size()];
          Short val;
          for(int i=0;i<processed.size();i++) {
             val = (Short) processed.elementAt(i);
             if(val!=null) ret[j++] = val.shortValue();
          }// end of for
          if(j!=processed.size()) {
             short[] buf = new short[j];
             for(int i=0;i<j;i++) buf[j] = ret[j];
             ret = buf;
          }// end of if
      }// end of if
      return ret;
    }// end of toShortArray

}// end of OwnershipManager

